import logging
from typing import Tuple
from google.protobuf.json_format import MessageToDict
import pandas as pd
import numpy as np
import os, psutil
import traceback
import uuid
import json
from datetime import datetime
import hashlib, json
import random
import simplejson

from abc import ABC, abstractclassmethod
import sqlalchemy as db
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.exc import IntegrityError, ProgrammingError
from sqlalchemy.sql.schema import MetaData, Table

from .query_templates import (
    create_serving_query,
    timestamp_template,
    get_index,
    get_time_index_field,
    join_decision_dict,
)
from .index_template import (
    create_training_query,
)

from .model_templates import create_model_view_query
from continual.services.dataingest.gcs_source import FeatureSourceCloudStore
from continual.python.sdk.exceptions import (
    InvalidArgumentError,
    FailedPreconditionError,
)
from continual.python.utils.utils import (
    create_limit_query,
    is_select_query,
    parse_feature_set_name,
)
from continual.rpc.management.v1 import management_types_pb2
from continual.python.sdk.query import Query
from .schema_utils import load_schemas, get_type_enum

# Map for continual typenames to the database types.
Typemap = {
    "FIXED": db.Float,
    "REAL": db.Float,
    "FLOAT": db.Float,
    "TEXT": db.Text,
    "DATE": db.Date,
    "DATETIME": db.DateTime,
    "TIMESTAMP": db.TIMESTAMP(),
    "VARIANT": db.Text,
    "TIMESTAMP_LTZ": db.TIMESTAMP(timezone=True),
    "TIMESTAMP_TZ": db.TIMESTAMP(timezone=True),
    "TIMESTAMP_NTZ": db.TIMESTAMP(),
    "OBJECT": db.Text,
    "STRING": db.Text,
    "ARRAY": db.ARRAY(db.Text),
    "BINARY": db.LargeBinary,
    "TIME": db.TIMESTAMP,
    "BOOLEAN": db.BOOLEAN,
    "BOOL": db.BOOLEAN,
    "NUMBER": db.Float,
    "FLOAT64": db.Float,
    "INTEGER": db.Integer,
    "INT": db.Integer,
    "INT64": db.Integer,
    "SMALLINT": db.Integer,
    "SMALLINT": db.Integer,
    "TINYINT": db.Integer,
    "NUMERIC": db.Float,
    "DECIMAL": db.Float,
    "BIGNUMERIC": db.Float,
    "BIGDECIMAL": db.Float,
}

featurestore_cache = {}


class FilterTables:
    def __init__(self, schema=None):
        self.schema = schema

    def __call__(self, table_name, meta_data):
        if "test_permissions_" in table_name:
            return False
        if self.schema is not None:
            if not table_name.startswith(f"{self.schema}."):
                return False
        return True


def join_conditions(left_id, right_id, left_ts, right_ts):
    des = join_decision_dict.get(
        (
            left_id is not None,
            left_ts is not None,
            right_id is not None,
            right_ts is not None,
        ),
        "invalid",
    )
    if des == "invalid":
        raise Exception("Invalid join condition")
    elif des == "ID":
        return left_id, right_id, None, None
    elif des == "TS":
        return left_ts, right_ts, None, None

    return left_ts, right_ts, left_id, right_id


class FeatureStore(ABC):
    def __init__(self, projectid=None):
        self.projectid = projectid
        self.featuresets = {}
        self.models = {}
        self.db_schema = None

    def _get_table_name(self, name):
        comps = parse_feature_set_name(name)
        if "projects" not in comps or (
            "featureSets" not in comps and "models" not in comps
        ):
            # TODO: Throw exception.
            pass

        table_name = ""
        if "featureSets" in comps:
            table_name = comps["featureSets"]
        else:
            table_name = comps["models"]

        project_name = comps["projects"]
        if "@" in project_name:
            project_name = project_name.replace("@", "_")
        return project_name, table_name

    def get_data_table_name(self, name):
        project, table_name = self._get_table_name(name)
        comps = parse_feature_set_name(name)
        if "models" in comps:
            table_name = "model_{}".format(table_name)
        if "featureSets" in comps:
            table_name = "featureset_{}".format(table_name)
        return project, table_name

    def get_table_name(self, schema):
        return self._get_table_name(schema.name)

    def get_ingesting_table_name(self, feature_set):
        proj, table_name = self.get_table_name(feature_set)
        return proj, table_name + "_ingesting"

    def get_model_prediction_table_name(self, model):
        proj, table_name = self.get_table_name(model)
        proj = proj.replace("@", "_")
        return proj, "model_{}_predictions_history".format(table_name)

    def get_model_view_name(self, model):
        proj, table_name = self.get_table_name(model)
        return proj, "model_{}_predictions".format(table_name)

    def create_project_schema(self):
        if not self.db_schema:
            raise InvalidArgumentError("schema_name is required")
        # try to see if the schema exists first
        try:
            print(f"Attempting to create project schema: {self.db_schema}")
            temp_id = random.randint(0, 10000)
            self.connection.execute(
                "CREATE TABLE IF NOT EXISTS {}.test_permissions_{} (id INTEGER)".format(
                    self.db_schema, temp_id
                )
            )
            self.connection.execute(
                "DROP TABLE IF EXISTS {}.test_permissions_{}".format(
                    self.db_schema, temp_id
                )
            )
        except Exception as e:
            logging.info(f"create schema exception: {e}")
            logging.info("schema does not appear to exist - create")
            self.connection.execute(
                "CREATE SCHEMA IF NOT EXISTS {}".format(self.db_schema)
            )

    def delete_project_schema(self, schema_name):
        self.connection.execute("DROP SCHEMA IF EXISTS {} CASCADE".format(schema_name))

    def create_feature_table(self, feature_set):
        """
         Creates featuretable.
         This does not check for the existance of the featuretable and should
           be done prior to calling this.

        Args: Feature set for which the tables are to be created.
        Returns None
        """
        if feature_set.schema.query is not None and feature_set.schema.query != "":
            self.create_featureset_view(feature_set)
            return

        _, table_name = self.get_table_name(feature_set)
        table_name = "featureset_{}".format(table_name)
        meta = db.MetaData(schema=self.db_schema)

        self.create_project_schema()

        columns = []

        for f in feature_set.schema.columns:
            columns.append(db.Column(f.name, Typemap[f.type]))

        db.Table(table_name, meta, *columns)

        # Following retry logic is added in case a parallel operation
        # creates a schema that might not have existed at the start of the session.
        count = 0
        done = False
        while not done:
            try:
                meta.create_all(self.engine)
                done = True
            except (IntegrityError, ProgrammingError):
                logging.debug("Error while creating feature table " + str(count))
                count = count + 1
                if count == 2:
                    done = True

    def create_model_tables(
        self, model, featureset_map, problem_type, prev_index, prev_time_index
    ):
        logging.info("create_model_tables: {}".format(model))
        if model.schema.query is not None and model.schema.query != "":
            self.create_model_data_view(model)

        # Create all the historical prediction tables.
        self.create_prediction_model_table(
            model, problem_type, prev_index, prev_time_index
        )
        self.create_model_view(model, featureset_map, True, problem_type)

    def set_db_schema(self, db_schema):
        self.db_schema = db_schema

    def create_model_view(self, model, fs_map, metadata, problem_type):
        """
        Create or refresh model historical prediction tables and a view that
          joins training data and predictions.
        """

        _, entity_name = self.get_data_table_name(model.name)
        table_name = model.schema.table
        if table_name is None or table_name == "":
            table_name = f"{self.db_schema}.{entity_name}"

        index = model.schema.index or None
        time_index = model.schema.time_index or None
        for f in model.schema.columns:
            if f.type == management_types_pb2.FieldType.INDEX:
                index = f.name
            elif f.type == management_types_pb2.FieldType.TIME_INDEX:
                time_index = f.name

        _, model_pred_name = self.get_model_prediction_table_name(model)
        proj, model_view_name = self.get_model_view_name(model)

        # The training data query returns qulified column names. in order to join this column name we need
        # to pass the name to be expected out of the traning query.
        index_match_name = index
        time_match_name = None
        if time_index:
            time_match_name = time_index

        model_query = create_model_view_query(
            table_name,
            index,
            time_index,
            index_match_name,
            time_match_name,
            model_pred_name,
            proj,
            model_view_name,
            model.schema.target,
            self.db_schema,
            problem_type,
        )
        logging.info(model_query)
        try:
            self.connection.execute(model_query)
        except StopIteration:
            pass
        except:
            logging.info(traceback.format_exc()[:3000])
            raise

    def create_prediction_model_table(
        self, model, problem_type, prev_index, prev_time_index
    ):
        """
        Create or update the table used to store predictions.

        This checks to see if the prediction table for this model exists, if it doesn't it creates it. If it does, it updates it if necessary.

        Parameters
        ----------
        model : continual.rpc.management.v1.Model
            The Model definition.
        problem_type : string
            The type of problem the model is solving.  E.g. regression, multiclass_classification, binary_classification.
        prev_index: string
            If there already is a model schema defined, the value of the previous model schema definition's index. Empty string if there is no previous model definition.
        prev_time_index: string
            If there already is a model schema defined, the value of the previous model schema definition's time index. Empty string if there is no previous model definition.
        """
        meta = db.MetaData(schema=self.db_schema)

        meta.reflect(bind=self.engine, only=FilterTables())
        _, model_pred_name = self.get_model_prediction_table_name(model)
        model_table = None
        for t in meta.sorted_tables:
            if model_pred_name == t.name:
                model_table = t
        if model_table is None:
            self._create_prediction_model_table(
                model, model_pred_name, meta, problem_type
            )
        else:
            self._update_prediction_model_table(
                model, model_table, problem_type, prev_index, prev_time_index
            )

    def _update_prediction_model_table(
        self, model, model_table: Table, problem_type, prev_index, prev_time_index
    ):

        logging.info(f"update table: {model_table.name}")
        index_column = model.schema.index
        time_index_column = model.schema.time_index
        logging.info("index_column: {} prev: {}".format(index_column, prev_index))
        logging.info(
            "time_index_column: {} prev: {}".format(time_index_column, prev_time_index)
        )

        target_column = model.schema.target
        prediction_col = f"{target_column}_prediction"
        prediction_details_col = f"{target_column}_prediction_details"
        prediction_score_col = f"{target_column}_prediction_score"
        if problem_type == "binary_classification":
            prediction_score_col = f"{target_column}_true_prediction_score"

        if (
            prediction_col in model_table.columns.keys()
            and (index_column == "" or index_column in model_table.columns.keys())
            and (
                time_index_column == ""
                or time_index_column in model_table.columns.keys()
            )
        ):
            # no changes to make
            logging.info("no changes to make to table")
            return
        # rename columns
        old_prediction_col = ""
        old_prediction_details_col = ""
        old_prediction_score_col = ""
        for col_name in model_table.columns.keys():
            if col_name.endswith("_prediction"):
                old_prediction_col = col_name
            if col_name.endswith("_prediction_details"):
                old_prediction_details_col = col_name
            if col_name.endswith("_prediction_score"):
                old_prediction_score_col = col_name

        if old_prediction_col == "":
            raise ValueError(
                f"predictions columms not found in table {model_table.name}"
            )

        with self.connection.begin():
            if old_prediction_col != prediction_col:
                self.connection.execute(
                    f"ALTER TABLE {self.db_schema}.{model_table.name} RENAME COLUMN {old_prediction_col} TO {prediction_col}"
                )
                if old_prediction_details_col != "":
                    self.connection.execute(
                        f"ALTER TABLE {self.db_schema}.{model_table.name} RENAME COLUMN {old_prediction_details_col} TO {prediction_details_col}"
                    )
                if old_prediction_score_col != "":
                    self.connection.execute(
                        f"ALTER TABLE {self.db_schema}.{model_table.name} RENAME COLUMN {old_prediction_score_col} TO {prediction_score_col}"
                    )

            if prev_index != index_column and index_column != "":
                self.connection.execute(
                    f"ALTER TABLE {self.db_schema}.{model_table.name} RENAME COLUMN {prev_index} TO {index_column}"
                )
            if prev_time_index != time_index_column and time_index_column != "":
                self.connection.execute(
                    f"ALTER TABLE {self.db_schema}.{model_table.name} RENAME COLUMN {prev_time_index} TO {time_index_column}"
                )

    def _create_prediction_model_table(
        self, model, model_pred_name, meta: MetaData, problem_type
    ):
        """
         Creates featuretable.
         This does not check for the existance of the featuretable and should
           be done prior to calling this.

        Args: Feature set for which the tables are to be created.≈
        Returns None
        """

        index = None
        time_index = None
        target_column = model.schema.target
        target_column_dtype = db.TEXT
        for f in model.schema.columns:
            if (
                f.type == management_types_pb2.FieldType.INDEX
                or f.name == model.schema.index
            ):
                index = f
            elif (
                f.type == management_types_pb2.FieldType.TIME_INDEX
                or f.name == model.schema.time_index
            ):
                time_index = f
            elif f.name == target_column:
                target_column_dtype = Typemap[f.dtype]

        columns = []
        columns.append(db.Column("batch_prediction", db.Text))
        columns.append(db.Column("model_version", db.Text))
        columns.append(db.Column("prediction_time", db.TIMESTAMP(timezone=True)))
        columns.append(db.Column(f"features", JSONB))

        if index is not None:
            columns.append(db.Column(index.name, Typemap[index.dtype]))

        if time_index is not None:
            columns.append(db.Column(time_index.name, Typemap[time_index.dtype]))
        columns.append(db.Column(f"{target_column}_prediction", target_column_dtype))
        if problem_type == "multiclass_classification":
            columns.append(db.Column(f"{target_column}_prediction_details", JSONB))
            columns.append(db.Column(f"{target_column}_prediction_score", db.Float))
        if problem_type == "binary_classification":
            columns.append(
                db.Column(f"{target_column}_true_prediction_score", db.Float)
            )

        db.Table(model_pred_name, meta, *columns)

        # Following retry logic is added in case a parallel operation
        # creates a schema that might not have existed at the start of the session.
        count = 0
        done = False
        while not done:
            try:
                meta.create_all(self.engine)
                done = True
            except (IntegrityError, ProgrammingError):
                logging.debug(traceback.format_exc())
                logging.debug("Error while creating model table " + str(count))
                count = count + 1
                if count == 2:
                    raise

    def drop_table(self, table_name, view_name):
        logging.info("drop_table: " + table_name + " view: " + view_name)

        try:
            self.connection.execute(
                "DROP VIEW IF EXISTS {}.{}".format(self.db_schema, view_name)
            )
        except:
            pass

        try:
            self.connection.execute(
                "DROP TABLE IF EXISTS {}.{}".format(self.db_schema, table_name)
            )
            return
        except StopIteration:
            pass
        except:
            logging.info(traceback.format_exc()[:3000])
            raise

    def load(self, dir_path):
        all_files = []
        if self.projectid is None:
            raise InvalidArgumentError("project needs to be set to load schemas")

        # r=root, d=directories, f = files
        for r, d, files in os.walk(dir_path):
            for f in files:
                if f.endswith(".yaml"):
                    all_files.append(os.path.join(r, f))

        count = 0
        for f in all_files:
            try:
                schema, schema_text = load_schemas(f)
                if schema is None:
                    print("Schema load failed")
                    continue

                if schema.query is not None and schema.query != "":
                    inferred, _ = self.infer_schema(schema.query, None, 0)
                    for col in schema.columns:
                        if col.name in inferred:
                            if col.name == "id" or col.name == schema.index:
                                col.type = management_types_pb2.FieldType.INDEX
                            # col.type = inferred[col.name]["type"]
                            col.dtype = inferred[col.name]["dtype"]
                            del inferred[col.name]

                    for col_name in inferred:
                        mc = management_types_pb2.ColumnConfig(
                            name=col_name,
                            type=get_type_enum(inferred[col_name].get("type", None)),
                            dtype=inferred[col_name]["dtype"],
                        )
                        schema.columns.append(mc)

                if schema.type == "FeatureSet":
                    fs = management_types_pb2.FeatureSet(
                        schema=schema, schema_text=schema_text
                    )
                    fs.name = (
                        "projects/" + self.projectid + "/featureSets/" + schema.name
                    )
                    self.featuresets[fs.name] = fs
                    if schema.table is None or schema.table == "":
                        self.create_feature_table(fs)

                    count = count + 1
                elif schema.type == "Model":
                    model = management_types_pb2.Model(
                        schema=schema, schema_text=schema_text
                    )
                    model.name = "projects/" + self.projectid + "/models/" + schema.name
                    self.models[model.name] = model
                    if (
                        schema.table is None
                        or schema.table == ""
                        and model.schema.target
                    ):
                        output_type = [
                            c.type
                            for c in model.schema.columns
                            if c.name == model.schema.target
                        ][0] or ""
                        problem_type = ""
                        if output_type == management_types_pb2.CATEGORICAL:
                            problem_type = "multiclass_classification"
                        if output_type == management_types_pb2.BOOLEAN:
                            problem_type = "binary_classification"
                        if output_type == management_types_pb2.NUMBER:
                            problem_type = "regression"
                        if output_type == management_types_pb2.TEXT:
                            problem_type = "text_prediction"
                        self.create_model_tables(
                            model, self.featuresets, problem_type, "", ""
                        )
                    else:
                        print("not creating model tables")
                    count = count + 1

            except Exception:
                raise
                # traceback.print_exc()

        return count

    def get_training_data(
        self,
        model,
        query=None,
        featureset_map=None,
        all_instances=False,
        latest=True,
        n=None,
        local=False,
    ):
        if featureset_map is None:
            featureset_map = self.featuresets

        if model is None:
            raise InvalidArgumentError("model required to get training data")

        if query is None:
            query = Query.form_query(
                model=model,
                all_featuresets=featureset_map,
            )

        df_iter = self.fetch_training_data(
            query.to_proto(),
            all_instances=all_instances,
            latest=latest,
            n=n,
            local=local,
        )

        return df_iter, query

    def create_featureset_view(self, feature_set):
        project, table_name = self.get_table_name(feature_set)
        table_name = "featureset_{}".format(table_name)
        query = feature_set.schema.query.strip().rstrip(";")
        create_query = "CREATE OR REPLACE VIEW {}.{} AS ({})".format(
            self.db_schema, table_name, query
        )
        logging.info(f"create_featureset_view query: {create_query}")
        self.connection.execute(create_query)

        return

    def create_model_data_view(self, model):
        _, table_name = self.get_data_table_name(model.name)
        query = model.schema.query.strip().rstrip(";")
        create_query = "CREATE OR REPLACE VIEW {}.{} AS ({})".format(
            self.db_schema, table_name, query
        )
        logging.info(f"create_model_data_view query: {create_query}")
        self.connection.execute(create_query)
        return

    def table_exists(self, schema_name, table_name):
        pass

    @abstractclassmethod
    def get_type(self) -> str:
        raise NotImplementedError()

    @abstractclassmethod
    def cleanup(self):
        raise NotImplementedError()

    @abstractclassmethod
    def is_hosted(self):
        raise NotImplementedError()

    @abstractclassmethod
    def close(self):
        raise NotImplementedError()

    @abstractclassmethod
    def get_stats_query_template(self):
        raise NotImplementedError()

    @abstractclassmethod
    def get_database_names(self):
        raise NotImplementedError()

    @abstractclassmethod
    def get_schema_names(self, database):
        raise NotImplementedError()

    @abstractclassmethod
    def get_table_names(self, database, schema):
        raise NotImplementedError()

    @abstractclassmethod
    def get_column_names(self, database, schema, table):
        raise NotImplementedError()

    @staticmethod
    def create_time_subquery(table_name, pkName, match_id, tsName, root_ts_name):
        params = {
            "EntityTable": table_name,
            "TsName": tsName,
            "PkName": pkName,
            "RootEntityID": match_id,
            "RootVisibilityTime": root_ts_name,
        }
        return timestamp_template.render(params=params)

    @staticmethod
    def create_featureset_query(
        name, db_schema, index, page_size=None, page_token=None, table_name=None
    ):
        parsed = parse_feature_set_name(name)
        if table_name is not None and table_name != "":
            full_name = table_name
        else:
            table_name = ""
            if "featureSets" in parsed:
                table_name = "featureset_" + parsed["featureSets"]
            else:
                table_name = "model_" + parsed["models"]

            full_name = db_schema + "." + table_name

        query = "SELECT * FROM {}".format(full_name)
        if page_token != None:
            if isinstance(page_token, str) and page_token != "":
                query = query + " WHERE {} > '{}'".format(str(index), str(page_token))
            else:
                query = query + " WHERE {} > {}".format(index, page_token)

        query = query + " ORDER BY {}".format(index)

        if page_size != None and page_size > 0:
            query = query + " LIMIT {}".format(page_size)

        return query

    def read_sql_query(
        self, query: str, chunksize: int = 1
    ) -> Tuple[pd.DataFrame, db.engine.ResultProxy]:
        """
        Slight rewrite of the Pandas SQL read_sql_query method. Instead of returning solely
        a DataFrame of the query results, this method will return the SQLAlchemy result proxy
        as well.
        """
        args = pd.io.sql._convert_params(query, None)

        pandas_sql = pd.io.sql.pandasSQL_builder(self.connection)
        result = pandas_sql.execute(*args)
        columns = result.keys()

        if chunksize is not None:
            return pandas_sql._query_iterator(result, chunksize, columns), result
        else:
            data = result.fetchall()
            frame = pd.io.sql._wrap_result(data, columns)
            return frame, result

    def infer_schema(
        self,
        query,
        excludes=None,
        sample_row_count=100,
        allow_invalid_types=False,
        timeout_ms=0,
    ):
        """
        Infers schema from the souce. Returns the schema as dictionary
         in following format.

        {
                 <feature_name> : <type>,
                 ...
             },
        """
        logging.info(f"infer_schema query: {query}")
        if is_select_query(query):
            query = create_limit_query(query, sample_row_count or 100)
        chunker, res = self.read_sql_query(query, chunksize=sample_row_count or 1)
        df = next(chunker)
        if excludes is not None:
            for c in excludes:
                if c in df.columns:
                    df = df.drop(c, axis=1)

        df.columns = [col.lower() for col in df.columns]
        inferred_schema = {}
        for col in df.select_dtypes(include=np.number).columns:
            colname = col.strip()
            if colname.startswith("unnammed:"):
                continue

            if len(colname) > 63:
                continue

            dtype = "FLOAT"
            if str(df[col].dtype).startswith("int"):
                dtype = "INTEGER"
            inferred_schema[colname] = {"type": "NUMBER", "dtype": dtype}

        for col in df.select_dtypes(include="object").columns:
            colname = col.strip()
            if colname.startswith("Unnammed:"):
                continue

            if len(colname) > 63:
                continue

            inferred_schema[colname] = {"type": "CATEGORICAL", "dtype": "TEXT"}

        for col in df.select_dtypes(include=["datetime", "datetimetz"]).columns:
            colname = col.strip()
            if colname.startswith("Unnammed:"):
                continue

            if len(colname) > 63:
                continue

            inferred_schema[colname] = {"type": "TIMESTAMP", "dtype": "TIMESTAMP"}

        for col in df.select_dtypes(include="bool").columns:
            colname = col.strip()
            if colname.startswith("Unnammed:"):
                continue

            if len(colname) > 63:
                continue

            inferred_schema[colname] = {"type": "BOOLEAN", "dtype": "BOOLEAN"}

        if allow_invalid_types and any(c not in inferred_schema for c in df.columns):
            type_codes = {c.name: c.type_code for c in res.cursor.description}
            types_df = pd.read_sql_query(
                "SELECT oid, typname FROM pg_type", self.connection
            )
            pg_types = {row[0]: row[1] for row in types_df.values}
            for col in df.columns:
                colname = col.strip()
                if colname.startswith("Unnammed:"):
                    continue
                if len(colname) > 63:
                    continue

                if colname not in inferred_schema and colname in type_codes:
                    inferred_schema[colname] = {
                        "type": "INVALID",
                        "dtype": pg_types.get(type_codes[colname], "INVALID"),
                    }

        all_records = []
        dicts = df.to_dict(orient="record")
        for d in dicts:
            all_records.append(simplejson.dumps(d, ignore_nan=True, default=str))
        return inferred_schema, all_records

    def fetch_feature_data(self, name, index, page_size, page_token, table_name):
        query = FeatureStore.create_featureset_query(
            name, self.db_schema, index, page_size, page_token, table_name
        )
        logging.info(query)
        return pd.read_sql_query(query, self.connection)

    def fetch_training_data(
        self,
        q,
        all_instances=False,
        n=None,
        local=False,
    ):
        if local:
            if isinstance(q, Query):
                qdict = q.to_dict()
            elif isinstance(q, management_types_pb2.Query):
                qdict = MessageToDict(q, preserving_proto_field_name=True)
            else:
                qdict = q

            return self.perform_local_join(qdict)

        """
        Returns a iterator for reading dataframes of traning data.
        """
        query, column_alias_map = create_training_query(
            q, self, all_instances=all_instances
        )
        logging.info(f"fetch_training_data query: {query}")
        if n == None:
            n = 100

        class CDFIter:
            def rename(self, df):
                df.rename(columns=self.mapping, inplace=True)
                return df

            def __init__(self, iterable, mapping):
                self.iterable = iterable
                self.mapping = mapping

            def __iter__(self):
                return self

            def __next__(self):
                ndf = next(self.iterable)
                return self.rename(ndf)

        return CDFIter(
            pd.read_sql_query(query, self.connection, chunksize=n), column_alias_map
        )

    def fetch_serving_data(self, q, overrides, data_store=None):
        """
        Returns a iterator for reading dataframes of traning data.
        """
        qd = q.override(overrides)
        if data_store is not None:
            data_store.get_type()
        query, params, sub_entities = create_serving_query(
            qd,
            False,
            self,
        )
        logging.info(query)
        df = pd.read_sql_query(query, self.connection)
        df.columns = df.columns.str.lower()
        if "finalrank" in df.columns:
            df = df.drop("finalrank", axis=1)

        for c in df.columns:
            if not c.endswith(".timerank"):
                continue
            df = df.drop(c, axis=1)

        return df

    def compute_feature_stats(self, name, columns, table_name=None):
        proj, entity_name = self.get_data_table_name(name)

        num_rows = 0
        features = []
        if table_name is None or table_name == "":
            table_name = f"{self.db_schema}.{entity_name}"

        res = self.connection.execute(f"SELECT count(*) as num_rows FROM {table_name}")
        row = next(res)
        num_rows = row[0]

        field_array = []  # TODO
        for f in columns:
            field_array.append(
                {
                    "name": f.name,
                    "type": f.dtype,
                    "ctype": management_types_pb2._FIELDTYPE.values_by_number[
                        f.type
                    ].name,
                }
            )

        params = {
            "AllFeatures": field_array,
            "FullName": table_name,
        }

        stats_query = self.get_stats_query_template().render(params=params)
        logging.info(stats_query)
        cur = self.connection.execute(stats_query)
        done = False
        computed_at = None
        while not done:
            try:
                row = next(cur)
                d = dict(zip(row.keys(), row))
                if "computed_at" in d:
                    computed_at = d["computed_at"]
                    del d["computed_at"]

                features.append(d)
            except StopIteration:
                done = True

        return num_rows, computed_at, features

    def load_data(self, filename, table_name, replace=False):
        source = FeatureSourceCloudStore(str(uuid.uuid4()), filename)
        df = source.fetch_all()
        df.columns = [col.lower() for col in df.columns]
        exists = "fail"
        if replace:
            exists = "replace"
        try:
            logging.info(
                "load_data: schema: {}, table_name: {}".format(
                    self.db_schema, table_name.lower()
                )
            )
            df.to_sql(
                table_name.lower(),
                self.connection,
                schema=self.db_schema,
                index=False,
                chunksize=10000,
                method="multi",
                if_exists=exists,
            )
        except ValueError:
            traceback.print_exc()

        return table_name.lower()

    def drop_featureset_view(self, name):
        project, table_name = self.get_data_table_name(name)
        self.connection.execute(
            "DROP VIEW IF EXISTS {}.{}".format(self.db_schema, table_name)
        )
        return

    def drop_model_views(self, name):
        project, table_name = self._get_table_name(name)
        view_name = "model_{}_predictions".format(table_name)
        self.connection.execute(
            "DROP VIEW IF EXISTS {}.{}".format(self.db_schema, view_name)
        )

        prediction_table_name = "model_{}_predictions_history".format(table_name)
        self.connection.execute(
            "DROP TABLE IF EXISTS {}.{}".format(self.db_schema, prediction_table_name)
        )

        project, table_name = self.get_data_table_name(name)
        self.connection.execute(
            "DROP VIEW IF EXISTS {}.{}".format(self.db_schema, table_name)
        )

        return

    def perform_local_join(self, q_dict):
        if q_dict is None:
            return None

        full_name = q_dict.get("table_name", None)

        project_id, table_name = self.get_data_table_name(q_dict["name"])
        if full_name is None:
            full_name = self.db_schema + "." + table_name

        df = pd.read_sql("select * from {}".format(full_name), self.engine)

        idx, qualified_idx = get_index(q_dict)
        time_idx, qualified_time_idx = get_time_index_field(q_dict)

        rename_dict = {}
        for field in q_dict["fields"]:
            rename_dict[field["name"]] = field["qualified_name"]
            if "type" in field and field["type"] == "TEXT":
                df[field["name"]] = df[field["name"]].astype("str")

            if "model" in field and field["model"]:
                df = df[df[field["name"]].notnull()]

            if "relation" in field and field["relation"] is not None:
                kdf = self.perform_local_join(field["relation"])
                r_idx, qualified_r_idx = get_index(field["relation"])
                r_time_idx, qualified_r_time_idx = get_time_index_field(
                    field["relation"]
                )

                logging.info(
                    "Join {} with {}".format(q_dict["name"], field["relation"]["name"])
                )
                left_on, right_on, left_by, right_by = join_conditions(
                    field["name"], qualified_r_idx, time_idx, qualified_r_time_idx
                )
                logging.info(
                    f"left_on {left_on}, right_on {right_on}, left_by {left_by}, right_by {right_by}"
                )
                df = df.sort_values(left_on)
                kdf = kdf.sort_values(right_on)
                if df.dtypes[left_on] != kdf.dtypes[right_on]:
                    raise FailedPreconditionError(
                        f"Unable to join data due to type mismatch: {left_on} {right_on}"
                    )

                if df.dtypes[left_on] == "O":
                    df = pd.merge(
                        df,
                        kdf,
                        left_on=left_on,
                        right_on=right_on,
                        suffixes=("", "_Continual"),
                        how="left",
                    )
                else:
                    df = pd.merge_asof(
                        df,
                        kdf,
                        left_on=left_on,
                        right_on=right_on,
                        left_by=left_by,
                        right_by=right_by,
                        suffixes=("", "_Continual"),
                        allow_exact_matches=True,
                    )
                del kdf

        if "related" in q_dict:
            for related in q_dict["related"]:
                kdf = self.perform_local_join(related)
                r_idx, qualified_r_idx = get_index(related)
                r_time_idx, qualified_r_time_idx = get_time_index_field(related)
                logging.info("Join {} with {}".format(q_dict["name"], related["name"]))
                left_on, right_on, left_by, right_by = join_conditions(
                    idx, qualified_r_idx, time_idx, qualified_r_time_idx
                )
                df = df.sort_values(left_on)
                kdf = kdf.sort_values(right_on)
                logging.info(
                    f"left_on {left_on}, right_on {right_on}, left_by {left_by}, right_by {right_by}"
                )
                if df.dtypes[left_on] != kdf.dtypes[right_on]:
                    raise FailedPreconditionError(
                        f"Unable to join data due to type mismatch: {left_on} {right_on}"
                    )

                if df.dtypes[left_on] == "O":
                    df = pd.merge(
                        df,
                        kdf,
                        left_on=left_on,
                        right_on=right_on,
                        suffixes=("", "_Continual"),
                        how="left",
                    )
                else:
                    df = pd.merge_asof(
                        df,
                        kdf,
                        left_on=left_on,
                        right_on=right_on,
                        left_by=left_by,
                        right_by=right_by,
                        suffixes=("", "_Continual"),
                        allow_exact_matches=True,
                    )
                del kdf

        for col in set(df.columns):
            if col.endswith("_Continual"):
                df.drop(col, axis=1, inplace=True)

        df.rename(rename_dict, inplace=True, axis=1)
        logging.info("Returning {}, {}".format(q_dict["name"], df.shape))
        process = psutil.Process(os.getpid())
        logging.info(
            "Current memory usage : {}".format(process.memory_info().rss / 1024 / 1024)
        )
        return df


def data_store_hash(datastore) -> str:
    """MD5 hash of a datastore."""
    if datastore is None:
        return "default"
    dictionary = MessageToDict(datastore)
    dhash = hashlib.md5()
    # We need to sort arguments so {'a': 1, 'b': 2} is
    # the same as {'b': 2, 'a': 1}
    encoded = json.dumps(dictionary, sort_keys=True).encode()
    dhash.update(encoded)
    return dhash.hexdigest()


def get_cached(data_store):
    key = data_store_hash(data_store)
    fs_entry = featurestore_cache.get(key, None)
    if fs_entry is None:
        return None

    if (datetime.now() - fs_entry["access_time"]).seconds > 3600:
        del featurestore_cache[key]
        return None

    return fs_entry["featurestore"]


def update_cache(data_store, featurestore):
    key = data_store_hash(data_store)
    fs_entry = featurestore_cache.get(key, None)
    if fs_entry is None:
        featurestore_cache[key] = {
            "featurestore": featurestore,
            "access_time": datetime.now(),
        }


def get_datastore_db_schema(data_store: management_types_pb2.DataStore = None):
    if data_store is not None:
        if data_store.type == "snowflake":
            return data_store.snowflake.db_schema
        elif data_store.type == "sqlserver":
            return data_store.sqlserver.db_schema
        elif data_store.type == "continual" or data_store.type == "postgres":
            return data_store.postgres.db_schema
        elif data_store.type == "bigquery":
            return data_store.big_query.dataset
        elif data_store.type == "redshift":
            return data_store.redshift.db_schema
        elif data_store.type == "databricks":
            return data_store.databricks.db_schema
    return ""


def get_featurestore(
    data_store: management_types_pb2.DataStore = None,
    cached: bool = False,
    cfg: dict = None,
    projectid=None,
    init=True,
) -> FeatureStore:
    """Get backing data store for feature store."""
    fs = None
    if cached:
        fs = get_cached(data_store)

    if fs is None:
        if data_store is not None and data_store.type != "":
            if data_store.type == "snowflake":
                from .snowflake import SnowflakeFeatureStore

                fs = SnowflakeFeatureStore(data_store, None, cfg)
            elif data_store.type == "sqlserver":
                from .mssqlserver import MSSqlFeatureStore

                fs = MSSqlFeatureStore(data_store, None, cfg)
            elif data_store.type == "continual" or data_store.type == "postgres":
                from .postgres import PostgresFeatureStore

                fs = PostgresFeatureStore(data_store, None, cfg)
            elif data_store.type == "bigquery":
                from .bigquery import BigqueryFeatureStore

                fs = BigqueryFeatureStore(data_store, None, cfg)
            elif data_store.type == "redshift":
                from .redshift import RedshiftFeatureStore

                fs = RedshiftFeatureStore(data_store, None, cfg)
            elif data_store.type == "databricks":
                from .databricks import DatabricksFeatureStore

                fs = DatabricksFeatureStore(data_store, None, cfg, revive_cluster=init)
        else:
            # for testing purposes only
            logging.info("initializing postgres featurestore for testing")
            from .postgres import PostgresFeatureStore

            fs = PostgresFeatureStore(data_store, None, cfg)

            # create dummy data store
            data_store = management_types_pb2.DataStore()
            pg = management_types_pb2.DataStore.Postgres()
            data_store.type = "continual"
            data_store.postgres.MergeFrom(pg)
            data_store.postgres.db_schema = projectid or ""

    if fs is not None:
        update_cache(data_store, fs)

    fs.projectid = projectid
    logging.debug(f"setting fs projectid to {projectid or 'empty'}, init: {init}")

    if get_datastore_db_schema(data_store):
        fs.set_db_schema(get_datastore_db_schema(data_store))
        if not fs.projectid:
            fs.projectid = get_datastore_db_schema(data_store)
        if init:
            fs.create_project_schema()

    logging.debug("featurestore projectid: {}".format(fs.projectid))

    return fs



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['continual',
 'continual.python',
 'continual.python.blob',
 'continual.python.cli',
 'continual.python.cli.tests',
 'continual.python.sdk',
 'continual.python.sdk.extensions',
 'continual.python.sdk.extensions.algorithms',
 'continual.python.sdk.extensions.models',
 'continual.python.sdk.extensions.templates',
 'continual.python.sdk.features',
 'continual.python.sdk.logger',
 'continual.python.utils',
 'continual.rpc',
 'continual.rpc.graphql',
 'continual.rpc.logging',
 'continual.rpc.logging.v1',
 'continual.rpc.management',
 'continual.rpc.management.v1',
 'continual.rpc.rpc',
 'continual.services',
 'continual.services.compute_services',
 'continual.services.compute_services.pipelines',
 'continual.services.compute_services.pipelines.batchprediction',
 'continual.services.compute_services.pipelines.training',
 'continual.services.compute_services.pipelines.training.tests',
 'continual.services.compute_services.utils',
 'continual.services.dataingest']

package_data = \
{'': ['*'],
 'continual.python': ['examples/bank_marketing/*',
                      'examples/kickstarter/*',
                      'extras/*']}

install_requires = \
['GitPython>=3.1.27',
 'click==8.0.4',
 'cloudpickle>=2.0.0',
 'cron-descriptor>=1.2.24',
 'databricks_api>=0.8.0',
 'dill>=0.3.4',
 'fsspec==2022.3',
 'gitpython>=3.1.7',
 'google-cloud-storage>=1.33.0',
 'grpcio>=1.27.1',
 'grpcio_status>=1.31.0',
 'halo>=0.0.30',
 'humanize>=2.5.0',
 'jinja2>=3.1.1',
 'mlflow>=1.23.1',
 'numpy>=1.21.6',
 'pandas-gbq>=0.14.1',
 'pandas>=1.0.1',
 'poetry>=0.12',
 'protobuf>=3.12.0,<3.19.0',
 'psutil>=5.8.0',
 'pyhive>=0.6.4',
 'pytz>=2020.5',
 'pyyaml>=5.4',
 'requests>=2.23.0',
 'rich>=9.13.0',
 'scikit_learn>=1.1.2',
 'shortuuid>=1.0.9',
 'simplejson>=3.17.6',
 'snowflake-sqlalchemy>=1.3.3',
 'sqlalchemy-bigquery>=1.4.3',
 'sqlalchemy-redshift>=0.8.9',
 'sqlalchemy>=1.4.27',
 'sqlparse>=0.4.2',
 'tableschema>=1.20.2',
 'tabulate>=0.8.6',
 'toml>=0.10.2',
 'tqdm>=4.54.1',
 'typer==0.4.0',
 'yamale>=4.0.0']

entry_points = \
{'console_scripts': ['continual = continual.python.cli.cli:cli']}

setup_kwargs = {
    'name': 'continual',
    'version': '0.5.81',
    'description': 'Operational AI for the Modern Data Stack',
    'long_description': '# Python CLI and SDK for Continual\n\nContinual is an operational AI for the modern data stack. Learn more at\nhttps://continual.ai.\n\n## Getting Started\n\nTo install the Continual CLI and SDK run:\n\n```\npip3 install continual\n```\n',
    'author': 'Continual',
    'author_email': 'support@continual.ai',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



from math import isinf
import mimetypes
from typing import Optional, Any, List, Tuple

from threading import Thread

import io
import os
import time
import grpc
import json
import requests

from google.resumable_media import DataCorruption
from google.resumable_media.requests import ResumableUpload

import numpy as np
import tarfile

from continual.python.utils.client_utils import get_management_channel

from continual.python.sdk.exceptions import normalize_exceptions_for_class
from continual.python.sdk.interceptors import AuthInterceptor

from continual.rpc.management.v1 import (
    management_types_pb2,
    management_pb2_grpc,
    management_pb2,
    types as management_types_py,
)
from continual.rpc.logging.v1 import (
    logging_pb2_grpc,
    logging_pb2,
)

CHUNK_SIZE = 117440512  # 112 MB in bytes, chosen arbitrarily


class NpEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.bool_):
            return bool(obj)
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


class Logger:
    def __init__(
        self,
        resource: str = "",
        session: str = "",
        endpoint: str = "",
        heartbeat_interval: Optional[int] = 10,
        disable_heartbeat: Optional[bool] = False,
    ):
        self._resource = resource
        self._session = session
        self._endpoint = endpoint
        self._disable_heartbeat = disable_heartbeat
        self._heartbeat_interval = heartbeat_interval

        self.channel: grpc.Channel = None
        self.client: logging_pb2_grpc.LoggingAPIStub = None
        self.mgmt_client: management_pb2_grpc.ManagementAPIStub = None

        if self._endpoint and self._session:
            auth_interceptor = AuthInterceptor(lambda: self.session)

            self.channel = get_management_channel(self.endpoint)

            self.channel = grpc.intercept_channel(self.channel, auth_interceptor)
            self.client = normalize_exceptions_for_class(
                logging_pb2_grpc.LoggingAPIStub(self.channel)
            )
            self.mgmt_client = normalize_exceptions_for_class(
                management_pb2_grpc.ManagementAPIStub(self.channel)
            )

        self.hb_thread: HeartbeatThread = None
        self.start_heartbeat_thread()

    def __del__(self) -> None:
        if self.hb_thread:
            self.hb_thread.stop()

    def init(self, resource: str = "", session: str = "", endpoint: str = ""):
        self.set_resource(resource)
        if session:
            self.set_session(session)
        if endpoint:
            self.set_endpoint(endpoint)

    def start_heartbeat_thread(self):
        if self.hb_thread:
            self.hb_thread.stop()
        self.hb_thread = HeartbeatThread(logger=self, daemon=True)
        self.hb_thread.start()

    def start_clients(self):
        if self.endpoint and self.session:
            auth_interceptor = AuthInterceptor(lambda: self.session)

            self.channel = get_management_channel(self.endpoint)

            self.channel = grpc.intercept_channel(self.channel, auth_interceptor)
            self.client = normalize_exceptions_for_class(
                logging_pb2_grpc.LoggingAPIStub(self.channel)
            )
            self.mgmt_client = normalize_exceptions_for_class(
                management_pb2_grpc.ManagementAPIStub(self.channel)
            )

    @property
    def resource(self) -> str:
        return self._resource

    def set_resource(self, r: str):
        if (not is_model_version(r)) and (not is_batch_prediction(r)):
            raise ValueError(
                f"Unrecognized resource name passed into logger: {r}. Must match name pattern of either ModelVersion or BatchPrediction."
            )
        restart_hb = r != self._resource
        self._resource = r
        if restart_hb:
            self.start_heartbeat_thread()

    @property
    def is_model_version(self) -> bool:
        return is_model_version(self.resource)

    @property
    def is_batch_prediction(self) -> bool:
        return is_batch_prediction(self.resource)

    @property
    def session(self) -> str:
        return self._session

    def set_session(self, s: str):
        if not is_session(s):
            raise ValueError(f"Invalid session token passed into logger: {s}")
        self._session = s
        self.start_clients()

    @property
    def endpoint(self) -> str:
        return self._endpoint

    def set_endpoint(self, e: str):
        if e not in (
            "https://api.continual.ai",
            "https://api-benchmark.continual.ai",
            "http://127.0.0.1",
            "http://localhost",
            "http://host.docker.internal",
            "http://cluster",
        ):
            raise ValueError(f"Unsupported endpoint passed into logger: {e}")
        self._endpoint = e
        self.start_clients()

    @property
    def disable_heartbeat(self) -> bool:
        return self._disable_heartbeat

    def set_disable_heartbeat(self, disable: bool):
        self._disable_heartbeat = disable

    @property
    def heartbeat_interval(self) -> int:
        return self._heartbeat_interval

    def set_heartbeat_interval(self, interval: int):
        if interval < 0 or isinf(interval):
            raise ValueError(
                f"Invalid heartbeat interval passed into logger: {interval}"
            )
        self._heartbeat_interval = interval

    def log_heartbeat(self):
        if not self.client or not self.resource:
            return
        req = logging_pb2.LogHeartbeatRequest(resource=self.resource)
        self.client.LogHeartbeat(req)
        # print(
        #     f"[HEARTBEAT] Resource=({self.resource}) at {datetime.utcnow().isoformat()}"
        # )

    def log_state(self, state: management_types_py.ModelVersionState):
        if not self.client or not self.is_model_version:
            print(f"State: {state}")
            return

        req = logging_pb2.LogModelVersionRequest(
            model_version=management_types_pb2.ModelVersion(
                name=self.resource, state=state.to_proto()
            ),
            update_paths=["state"],
        )
        res = self.client.LogModelVersion(req)
        print(f"log_state: {res}")

    def log_field(self, field_name: str, value: Any):
        if not self.client:
            raise Exception(f"Client not set in logger.")

        resource_args = {"name": self.resource, field_name: value}

        if self.is_model_version:
            req = logging_pb2.LogModelVersionRequest(
                model_version=management_types_pb2.ModelVersion(**resource_args),
                update_paths=[field_name],
            )
            res = self.client.LogModelVersion(req)
        else:
            req = logging_pb2.LogBatchPredictionRequest(
                batch_prediction=management_types_pb2.BatchPrediction(**resource_args),
                update_paths=[field_name],
            )
            res = self.client.LogBatchPrediction(req)
        print(f"log_field: {field_name} with {value} on {self.resource} : {res}")

    def log_predict_state(self, state: management_types_py.BatchPredictionState):
        if not self.client or not self.is_batch_prediction:
            print(f"State: {state}")
            return

        req = logging_pb2.LogBatchPredictionRequest(
            batch_prediction=management_types_pb2.BatchPrediction(
                name=self.resource, state=state.to_proto()
            ),
            update_paths=["state"],
        )
        res = self.client.LogBatchPrediction(req)
        print(f"log_predict_state: {res}")

    def get_artifact(self, name: str) -> management_types_pb2.Artifact:
        if not self.client:
            print(f"Cannot fetch artifact without client")
            return

        req = logging_pb2.GetArtifactRequest(name=name)
        res = self.client.GetArtifact(req)
        return res

    def delete_artifact(self, name: str):
        if not self.client:
            print(f"Cannot delete artifact without client")
            return

        req = logging_pb2.DeleteArtifactRequest(name=name)
        self.client.DeleteArtifact(req)

    def get_model_artifact(
        self,
        model_version: str = "",
        download: bool = False,
        download_dir: str = "./artifacts",
    ) -> Tuple[management_types_pb2.Artifact, str]:
        if not self.client:
            print(f"Cannot fetch artifact without client")
            return

        req = management_pb2.GetModelVersionRequest(name=model_version or self.resource)
        res = self.mgmt_client.GetModelVersion(req)
        print(f"get_model_artifact res: {res}")
        if not res.model_artifact:
            raise ValueError(
                f"Model version ({model_version}) does not have a model artifact."
            )

        artifact = self.get_artifact(res.model_artifact)
        downloaded_to = ""
        if download:
            if not artifact.url:
                raise ValueError(
                    f"Artifact cannot be downloaded - no URL was found: {artifact.url}"
                )
            try:
                res = requests.get(artifact.url, stream=True)
                with tarfile.open(fileobj=res.raw, mode="r") as f:
                    f.extractall(download_dir)

                root_dir = os.path.basename(artifact.path or "").split(".")[0]
                downloaded_to = os.path.join(download_dir, root_dir)
            except:
                res = requests.get(artifact.url)
                downloaded_to = os.path.join(
                    download_dir,
                    os.path.basename(artifact.path or artifact.name.split("/")[-1]),
                )
                with open(downloaded_to, "wb") as f:
                    f.write(res.content)
        return artifact, downloaded_to

    def log_model_artifact(self, path: str, metadata: dict = None) -> str:
        if not self.is_model_version:
            print(f"Cannot log model artifact to a BatchPrediction.")
            return

        artifact_name = self.log_artifact(path=path, type="model", metadata=metadata)
        req = logging_pb2.LogModelVersionRequest(
            model_version=management_types_pb2.ModelVersion(
                name=self.resource,
                model_artifact=artifact_name,
            ),
            update_paths=["model_artifact"],
        )
        res = self.client.LogModelVersion(req)
        print(f"log_model_artifact: {res}")
        return artifact_name

    def log_artifacts(
        self, paths: List[str], type: str = "", metadata: dict = None
    ) -> List[str]:
        artifact_names = []
        for path in paths:
            try:
                name = self.log_artifact(path=path, type=type, metadata=metadata)
                artifact_names.append(name)
            except:
                artifact_names.append("")
        return artifact_names

    def log_artifact(self, path: str, type: str = "", metadata: dict = None) -> str:
        if not metadata:
            metadata = {}

        if not self.client:
            print(f"Artifact: path={path}, type={type} metadata={metadata}")
            return

        if not os.path.exists(path):
            raise ValueError(
                f"Invalid file path provided. Path ({path}) does not exist."
            )

        artifact_name = ""

        try:
            file_to_upload = path
            if os.path.isdir(path):
                tarfile_name = os.path.basename(path) + ".tar.gz"
                with tarfile.open(tarfile_name, "w:gz") as tar:
                    tar.add(
                        path,
                        recursive=True,
                        arcname=os.path.basename(tarfile_name).split(".")[0],
                    )
                file_to_upload = tarfile_name

            mime_type, _ = mimetypes.guess_type(file_to_upload)

            payload = ""
            with open(file_to_upload, "rb") as f:
                payload = f.read()

            total_bytes = len(payload)
            exceeds_chunk_size = total_bytes >= CHUNK_SIZE

            req = logging_pb2.GenerateArtifactUploadURLRequest(
                resource=self.resource,
                path=path,
                type=type,
                mime_type=mime_type or "",
                metadata=json.dumps(metadata or dict()),
                resumable=exceeds_chunk_size,
            )
            res = self.client.GenerateArtifactUploadURL(req)

            upload_url = res.url
            artifact_name = res.artifact.name

            headers = dict()
            if mime_type:
                headers["Content-Type"] = mime_type

            if exceeds_chunk_size:
                transport = requests.Session()
                metadata = dict(name=artifact_name, file=path)

                stream = io.BytesIO(payload)

                upload = ResumableUpload(upload_url, CHUNK_SIZE)
                upload._resumable_url = upload_url
                upload._total_bytes = total_bytes
                upload._stream = stream

                while not upload.finished:
                    try:
                        print("Writing chunk ... ")
                        response = upload.transmit_next_chunk(transport, timeout=60)
                        print(f"Chunk response: {response}")
                    except DataCorruption:
                        raise
            else:
                upload_res = requests.put(
                    upload_url, data=payload, headers={"Content-Type": mime_type}
                )
                upload_res.raise_for_status()

            return artifact_name
        except:
            if artifact_name:
                req = logging_pb2.DeleteArtifactRequest(name=artifact_name)
                res = self.client.DeleteArtifact(req)
            raise

    def log_data_checks(self, data_checks: management_types_pb2.DataChecks):
        if not self.client:
            print(f"Data Checks: {data_checks}")
            return

        req = logging_pb2.LogDataChecksRequest(
            resource=self.resource, data_checks=data_checks
        )
        res = self.client.LogDataChecks(req)
        print(f"log_data_checks: {res}")

    def log_prediction_stats(self, prediction_stats: management_types_pb2.DatasetStats):
        if not self.client or not self.is_batch_prediction:
            print(f"Prediction Stats: {prediction_stats}")
            return

        req = logging_pb2.LogPredictionStatsRequest(
            batch_prediction=self.resource, prediction_stats=prediction_stats
        )
        res = self.client.LogPredictionStats(req)
        print(f"log_prediction_stats: {res}")

    def log_eval_model_version_metrics(
        self,
        eval_model_version_metrics: management_types_pb2.EvaluationModelVersionMetrics,
    ):
        if not self.client:
            print(f"Dataset Stats: {eval_model_version_metrics}")
            return

        req = logging_pb2.LogEvaluationModelMetricsRequest(
            model_version=self.resource,
            evaluation_model_version_metrics=eval_model_version_metrics,
        )
        res = self.client.LogEvaluationModelMetrics(req)
        print(f"log_eval_model_version_metrics: {res}")

    def log_dataset_stats(self, dataset_stats: management_types_pb2.DatasetStats):
        if not self.client:
            print(f"Dataset Stats: {dataset_stats}")
            return

        req = logging_pb2.LogDatasetStatsRequest(
            resource=self.resource, dataset_stats=dataset_stats
        )
        res = self.client.LogDatasetStats(req)
        print(f"log_dataset_stats: {res}")

    def log_metrics(self, metrics: List[management_types_py.TrainingMetrics]):
        if not self.client:
            print(f"Metrics: {metrics}")
            return

        req = logging_pb2.LogMetricsRequest(
            model_version=self.resource,
            metrics=list(map(lambda m: m.to_proto(), metrics)),
        )
        res = self.client.LogMetrics(req)
        print(f"log_metrics: {res}")

    def log_experiment(self, experiment: management_types_py.Experiment):
        self.log_experiments(experiments=[experiment])

    def log_experiments(self, experiments: List[management_types_py.Experiment]):
        if not self.client:
            print(f"Experiments: {experiments}")
            return
        req = logging_pb2.LogExperimentsRequest(
            model_version=self.resource,
            experiments=list(map(lambda e: e.to_proto(), experiments)),
        )
        res = self.client.LogExperiments(req)
        print(f"log_experiments: {res}")

    def log_metadata(self, metadata: dict = None):
        if not self.client:
            print(f"Metadata: {metadata}")
            return
        self.log_field("metadata", json.dumps(metadata, cls=NpEncoder))


class HeartbeatThread(Thread):
    def __init__(self, logger: Logger, *args, **kwargs) -> None:
        Thread.__init__(self, *args, **kwargs)
        self.logger = logger
        self.running = True
        self.should_stop = False

    def stop(self):
        self.should_stop = True
        self.running = False

    def run(self) -> None:
        while True:
            if self.should_stop:
                break
            self.logger.log_heartbeat()
            time.sleep(self.logger.heartbeat_interval or 5)


def is_model_version(name: str) -> bool:
    name_split = name.split("/")
    if (
        len(name_split) != 6
        or name_split[0] != "projects"
        or name_split[2] != "models"
        or name_split[4] != "versions"
    ):
        return False
    return True


def is_batch_prediction(name: str) -> bool:
    name_split = name.split("/")
    if (
        len(name_split) != 6
        or name_split[0] != "projects"
        or name_split[2] != "models"
        or name_split[4] != "batchPredictions"
    ):
        return False
    return True


def is_session(name: str) -> bool:
    name_split = name.split("/")
    if len(name_split) != 2 or name_split[0] != "sessions":
        return False
    return True


logger = Logger()



import os
import yaml
import time
import pytest
import shortuuid

from continual.rpc.management.v1 import management_types_pb2
from continual.rpc.dataingest.v1 import dataingest_pb2

from continual.python.sdk.features.featurestore import get_featurestore
from continual.services.dataingest.server import DataIngestService

TEST_CONFIG_FILE = os.path.join(os.path.expanduser("~"), ".continual/testcfg.yaml")


@pytest.fixture(scope="session")
def postgres_mgmt():
    os.environ["CONFIG_ENV"] = "test"
    project_id = "testproject_" + shortuuid.uuid().lower()
    fstore = get_featurestore(projectid=project_id)
    fstore.cleanup()
    fstore.create_project_schema()
    fstore.load_data("gs://continual-dev/testing_data/train1.csv", project_id, "tweets")
    fstore.load_data(
        "gs://continual-dev/testing_data/username.csv", project_id, "twitter_user_raw"
    )
    fstore.load_data(
        "gs://continual-dev/testing_data/bank_train.csv", project_id, "bank_train"
    )

    fstore.load("./continual/services/dataingest/testdata/v2")

    mgmt = ManagementDummy(feature_store=fstore, project_id=project_id)
    print("Postgres featurestore ready")
    time.sleep(2)
    return mgmt


@pytest.fixture(scope="session")
def snowflake_mgmt():
    sf_datastore = get_snowflake_datastore()
    mgmt_snowflake = None
    if sf_datastore is None:
        return None

    project_id = "testproject_" + shortuuid.uuid().lower()
    if not sf_datastore.snowflake.db_schema:
        sf_datastore.snowflake.db_schema = project_id
    fstore = get_featurestore(sf_datastore, projectid=project_id)
    fstore.cleanup()
    fstore.create_project_schema()
    fstore.load_data(
        "gs://continual-dev/testing_data/train1.csv", project_id, "tweets_raw"
    )
    fstore.load_data(
        "gs://continual-dev/testing_data/username.csv", project_id, "twitter_user_raw"
    )
    fstore.load_data(
        "gs://continual-dev/testing_data/bank_train.csv", project_id, "bank_train"
    )
    fstore.load("./examples/twitter/v2")
    fstore.load("./examples/bank_marketing/v2")
    mgmt_snowflake = ManagementDummy(feature_store=fstore, project_id=project_id)

    print("Snowflake featurestore ready")
    return mgmt_snowflake


@pytest.fixture(scope="session")
def bigquery_mgmt():
    bq_datastore = get_bigquery_datastore()
    mgmt_bq = None
    if bq_datastore is None:
        return None

    project_id = "testproject_" + shortuuid.uuid().lower()
    fstore = get_featurestore(bq_datastore, projectid=project_id)
    fstore.cleanup()
    fstore.create_project_schema()
    fstore.load_data(
        "gs://continual-dev/testing_data/train1.csv", project_id, "tweets_raw"
    )
    fstore.load_data(
        "gs://continual-dev/testing_data/username.csv", project_id, "twitter_user_raw"
    )
    fstore.load_data(
        "gs://continual-dev/testing_data/bank_train.csv", project_id, "bank_train"
    )
    fstore.load("./examples/twitter/bigquery")
    fstore.load("./examples/bank_marketing/bigquery")
    mgmt_bq = ManagementDummy(feature_store=fstore, project_id=project_id)

    print("Snowflake featurestore ready")
    return mgmt_bq


@pytest.fixture
def mssql_mgmt():
    ms_datastore = get_mssql_datastore()
    mgmt_sqlserver = None
    if ms_datastore is None:
        return None

    project_id = "testproject_" + shortuuid.uuid().lower()
    fstore = get_featurestore(ms_datastore, projectid=project_id)
    fstore.cleanup()
    fstore.load("./examples/twitter/v2")
    mgmt_sqlserver = ManagementDummy(feature_store=fstore, project_id=project_id)
    print("Mssql featurestore ready")
    return mgmt_sqlserver


@pytest.fixture(scope="session")
def databricks_mgmt():
    print("Going to databricks management")
    db_datastore = get_databricks_datastore()
    if db_datastore is None:
        return None

    project_id = "testproject_" + shortuuid.uuid().lower()
    if not db_datastore.databricks.db_schema:
        db_datastore.databricks.db_schema = project_id
    fstore = get_featurestore(db_datastore, projectid=project_id)
    fstore.cleanup()
    fstore.create_project_schema()
    fstore.load_data("gs://continual-dev/testing_data/train1.csv", "tweets_raw")
    fstore.load_data("gs://continual-dev/testing_data/username.csv", "twitter_user_raw")
    fstore.load_data("gs://continual-dev/testing_data/bank_train.csv", "bank_train")

    # Note: The redshift examples work for databricks
    fstore.load("./examples/bank_marketing/redshift")
    mgmt_databricks = ManagementDummy(feature_store=fstore, project_id=project_id)

    print("Databricks featurestore ready")
    return mgmt_databricks


@pytest.fixture(scope="session")
def redshift_mgmt():
    print("Going to create redshift management")
    rs_datastore = get_redshift_datastore()
    mgmt_redshift = None
    if rs_datastore is None:
        return None

    project_id = "testproject_" + shortuuid.uuid().lower()
    fstore = get_featurestore(rs_datastore, projectid=project_id)
    fstore.cleanup()
    fstore.create_project_schema()
    fstore.load_data(
        "gs://continual-dev/testing_data/train1.csv", project_id, "tweets_raw"
    )
    fstore.load_data(
        "gs://continual-dev/testing_data/username.csv", project_id, "twitter_user_raw"
    )
    fstore.load_data(
        "gs://continual-dev/testing_data/bank_train.csv", project_id, "bank_train"
    )

    # Note: The bigquery examples work for redshift
    fstore.load("./examples/twitter/bigquery")
    fstore.load("./examples/bank_marketing/redshift")
    mgmt_redshift = ManagementDummy(feature_store=fstore, project_id=project_id)

    print("Redshift featurestore ready")
    return mgmt_redshift


def get_type_enum(typename):

    if typename is None:
        return None

    return management_types_pb2._FIELDTYPE.values_by_name[typename.upper()].number


def get_snowflake_datastore():
    if not os.path.isfile(TEST_CONFIG_FILE):
        print(
            "Test config file not found. Snowflake datastore not created ",
            __file__,
        )
        return None

    with open(TEST_CONFIG_FILE) as inp_file:
        cfg = yaml.load(inp_file, Loader=yaml.FullLoader)
        if "snowflake_connection" not in cfg:
            print(
                "Snowflake connection info not found in test configuration file, skipping ",
                __file__,
            )
            return None

        connection_info = cfg["snowflake_connection"]

        d = management_types_pb2.DataStore(
            type=connection_info["type"],
            snowflake=management_types_pb2.DataStore.Snowflake(
                username=connection_info["username"],
                database=connection_info["database"],
                password=connection_info["password"],
                host=connection_info["host"],
                warehouse=connection_info["warehouse"],
            ),
        )

        return d


def get_databricks_datastore():
    if not os.path.isfile(TEST_CONFIG_FILE):
        print(
            "Test config file not found. databricks datastore not created ",
            __file__,
        )
        return None

    with open(TEST_CONFIG_FILE) as inp_file:
        cfg = yaml.load(inp_file, Loader=yaml.FullLoader)
        if "databricks_connection" not in cfg:
            print(
                "Databricks connection info not found in test configuration file, skipping ",
                __file__,
            )
            return None

        connection_info = cfg["databricks_connection"]
        print(f"connection_info: {connection_info}")

        d = management_types_pb2.DataStore(
            type=connection_info["type"],
            databricks=management_types_pb2.DataStore.Databricks(
                http_path=connection_info["http_path"],
                database=connection_info["database"],
                token=connection_info["token"],
                host=connection_info["host"],
                port=connection_info["port"],
            ),
        )

        return d


def get_redshift_datastore():
    if not os.path.isfile(TEST_CONFIG_FILE):
        print(
            "Test config file not found. redshift datastore not created ",
            __file__,
        )
        return None

    with open(TEST_CONFIG_FILE) as inp_file:
        cfg = yaml.load(inp_file, Loader=yaml.FullLoader)
        if "redshift_connection" not in cfg:
            print(
                "Redshift connection info not found in test configuration file, skipping ",
                __file__,
            )
            return None

        connection_info = cfg["redshift_connection"]

        d = management_types_pb2.DataStore(
            type=connection_info["type"],
            redshift=management_types_pb2.DataStore.Redshift(
                username=connection_info["username"],
                database=connection_info["database"],
                password=connection_info["password"],
                host=connection_info["host"],
                port=connection_info["port"],
            ),
        )

        return d


def get_bigquery_datastore():
    if not os.path.isfile(TEST_CONFIG_FILE):
        print(
            "Test config file not found. Snowflake datastore not created ",
            __file__,
        )
        return None

    with open(TEST_CONFIG_FILE) as inp_file:
        cfg = yaml.load(inp_file, Loader=yaml.FullLoader)
        if "bigquery_credentials" not in cfg:
            print(
                "BigQuery credentials not found in test configuration file, skipping ",
                __file__,
            )
            return None

        connection_info = cfg["bigquery_credentials"]
        data = None
        with open(connection_info["cred_file"], "r") as fp:
            data = fp.read()

        d = management_types_pb2.DataStore(
            type="bigquery",
            big_query=management_types_pb2.DataStore.BigQuery(auth_file=data),
        )
        return d


def get_mssql_datastore():
    if not os.path.isfile(TEST_CONFIG_FILE):
        print(
            "Test config file not found. mssql datastore not created ",
            __file__,
        )
        return None

    with open(TEST_CONFIG_FILE) as inp_file:
        cfg = yaml.load(inp_file, Loader=yaml.FullLoader)
        if "mssql_connection" not in cfg:
            print(
                "Mssql connection info not found in test configuration file, skipping ",
                __file__,
            )
            pytest.skip(
                "Mssql connection info not found in test configuration file, skipping "
            )

        connection_info = cfg["mssql_connection"]

        d = management_types_pb2.DataStore(
            type=connection_info["type"],
            username=connection_info["username"],
            database=connection_info["database"],
            password=connection_info["password"],
            host=connection_info["host"],
            port=connection_info["port"],
        )
        return d


class ManagementDummy:
    def __init__(self, **kwargs):
        self.feature_sets = {}
        self.fs_store = kwargs["feature_store"]
        self.project_id = kwargs["project_id"]
        self.project_name = "projects/" + self.project_id

    def _get_latest_fs_name(self, name):
        latest_version = 0
        latest_version_name = None

        check_name = name.split("@")[0]
        for fs_name in self.feature_sets.keys():
            if fs_name.startswith(check_name):
                version = fs_name.split("@")[1]
                if version > latest_version:
                    latest_version = version
                    latest_version_name = fs_name

        return latest_version_name

    def PutFeatureSet(self, request):

        """
        Assignes a version and store the featureset.
        Currently only one version is supported.
        """
        featureSetname = request.parent + "/" + request.feature_set.schema.name
        request.feature_set.name = featureSetname

        fs = management_types_pb2.FeatureSet()
        fs.CopyFrom(request.feature_set)
        fs.name = featureSetname
        self.feature_sets[featureSetname] = fs

        self.fs_store.create_feature_table(fs)
        return request.feature_set

    def GetFeatureSet(self, request):

        """
        Returns the featureset with given name.
           TODO: Return the latest featureset if the version is not provided.
        """
        return self.feature_sets.get(request.name, None)

    def BatchGetFeatureSets(self, request):
        pass

    def GetFeatureSetMap(self):
        return self.feature_sets


def test_store(postgres_mgmt):
    featureset = management_types_pb2.FeatureSet()
    featureset.name = "projects/" + postgres_mgmt.project_id + "/featureSets/storetest"

    tab = postgres_mgmt.fs_store.get_feature_table(featureset)
    assert tab is None, "Feature table exists unexpectedely"


def create_project(mgmt):
    if mgmt is None:
        return

    print("create_project: " + mgmt.fs_store.get_type())
    project_id = "test123_" + shortuuid.uuid().lower()
    mgmt.fs_store.set_db_schema(project_id)

    mgmt.fs_store.create_project_schema()


def test_create_project(
    postgres_mgmt, snowflake_mgmt, bigquery_mgmt, redshift_mgmt, databricks_mgmt
):
    create_project(postgres_mgmt)
    create_project(snowflake_mgmt)
    create_project(bigquery_mgmt)
    create_project(redshift_mgmt)
    create_project(databricks_mgmt)


def compute_stats(mgmt):
    if mgmt is None:
        return
    print("compute_stats: " + mgmt.fs_store.get_type())

    name = "projects/" + mgmt.project_id + "/featureSets/bankmarketing"
    num_rows, computed_at, features = mgmt.fs_store.compute_feature_stats(
        name, mgmt.fs_store.featuresets[name].schema.columns
    )
    assert num_rows > 0, "compute stats failed"
    assert len(features) > 0, "compute stats failed"


def test_compute_stats(snowflake_mgmt, bigquery_mgmt, redshift_mgmt, databricks_mgmt):
    compute_stats(snowflake_mgmt)
    compute_stats(bigquery_mgmt)
    compute_stats(redshift_mgmt)
    compute_stats(databricks_mgmt)


def training_data(mgmt):
    if mgmt is None:
        return
    print("training_data: " + mgmt.fs_store.get_type())
    name = "projects/" + mgmt.project_id + "/models/bankmarketing_duration"

    dfi, query = mgmt.fs_store.get_training_data(
        mgmt.fs_store.models[name],
        featureset_map=mgmt.fs_store.featuresets,
        local=False,
        n=50_000,
    )
    try:
        df = next(dfi)
        print(df.shape)
        assert query, "get_training_data failed"
        assert df.shape[0] == 41188, "Unexpected number of rows returned"
    except StopIteration:
        assert False, "DId not get the dataframe rom the iterator"

    dfi, query = mgmt.fs_store.get_training_data(
        mgmt.fs_store.models[name],
        featureset_map=mgmt.fs_store.featuresets,
        n=10000,
        local=False,
    )
    done = False
    count = 0
    assert query, "get_training_data failed"
    while not done:
        try:
            df = next(dfi)
            assert df.shape[0] > 0, "Unexpected number of rows returned"
            assert df.shape[0] <= 10000, "Unexpected number of rows returned"
            count = count + 1
            assert count <= 5, "More iteration than expected"
        except:
            done = True

    assert count == 5, "More iteration than expected"


def test_get_training_data(
    snowflake_mgmt, bigquery_mgmt, redshift_mgmt, databricks_mgmt
):
    training_data(snowflake_mgmt)
    training_data(bigquery_mgmt)
    training_data(redshift_mgmt)
    training_data(databricks_mgmt)


def serving_data(mgmt):
    if mgmt is None:
        return
    print("serving_data: " + mgmt.fs_store.get_type())
    name = "projects/" + mgmt.project_id + "/models/bankmarketing_duration"

    dfi, query = mgmt.fs_store.get_training_data(
        mgmt.fs_store.models[name],
        featureset_map=mgmt.fs_store.featuresets,
        local=False,
    )
    df = next(dfi)
    assert query, "get_training_data failed"
    assert df.shape[0] == 41188, "Unexpected number of rows returned"

    data = mgmt.fs_store.fetch_serving_data(
        query,
        overrides={"id": "100"},
        data_store=mgmt.fs_store,
    )
    print(data.columns)
    assert data.shape[1] == 22, "Unexpected number of columns received"
    assert data.shape[0] == 1, "Unexpected number of columns received"

    """
    req = management_pb2.GetFeatureSetRequest(
        name="projects/" + mgmt.project_id + "/featureSets/tweet_rel"
    )
    mgmt.GetFeatureSet(req)
    all_featuresets = mgmt.GetFeatureSetMap()

    dfi, query = mgmt.fs_store.get_training_data("tweet_rel:rating", all_featuresets)
    df = next(dfi)
    assert query, "get_training_data failed"
    assert df.shape[0] == 999, "Unexpected number of rows returned"

    data = mgmt.fs_store.fetch_serving_data(
        query,
        overrides={"tweet_rel:id": "20"},
        data_store=mgmt.fs_store,
    )
    print(data)
    assert data.shape[1] == 10, "Unexpected number of columns received"
    assert data.shape[0] == 1, "Unexpected number of rows received"
    assert data["tweet_rel:id"][0] == "20"

    data = mgmt.fs_store.fetch_serving_data(
        query,
        overrides={"tweet_rel:id": "20", "tweet_rel:text": "Hope this works"},
        data_store=mgmt.fs_store,
    )
    assert data.shape[1] == 10, "Unexpected number of columns received"
    assert data.shape[0] == 1, "Unexpected number of rows received"
    assert data["tweet_rel:id"][0] == "20"
    assert data["tweet_rel:text"][0] == "Hope this works"

    data = mgmt.fs_store.fetch_serving_data(
        query,
        overrides={"tweet_rel:text": "Hope this works"},
        data_store=mgmt.fs_store,
    )
    assert data.shape[0] == 1, "Unexpected number of rows received"
    assert data["tweet_rel:text"][0] == "Hope this works"
    """


# def test_get_serving_data(snowflake_mgmt, bigquery_mgmt):
#    print("test_get_serving_data: ")
#    #serving_data(snowflake_mgmt)
#    serving_data(bigquery_mgmt)


def test_infer_schema_from_query_snowflake():
    data_store = get_snowflake_datastore()
    if data_store is None:
        print("Test configuration not found. skipping ")
        pytest.skip("Test configuration not found. skipping")

    env = management_types_pb2.Environment(
        name="projects/testproject", data_store=data_store
    )
    req = dataingest_pb2.GetInferredSchemaRequest(
        environment=env,
        query='SELECT id, TO_NUMBER(id, 6) as intid, text, rating  from  "CONTINUAL".featurestore_test.tweets order by intid',
    )

    os.environ["CONFIG_ENV"] = "local"
    server = DataIngestService()

    resp = server.GetInferredSchema(req, None)

    print("number of features found ", str(len(resp.columns)))
    print("sample data size", resp.record_count)

    assert len(resp.columns) == 4, "Unexpected number of features inferred"
    assert resp.record_count > 0, "No sample data returned"

    req = dataingest_pb2.GetInferredSchemaRequest(
        environment=env,
        query='SELECT id, TO_NUMBER(id, 6) as intid, text, rating  from  "CONTINUAL".featurestore_test.tweets order by intid',
    )

    resp = server.GetInferredSchema(req, None)

    assert len(resp.columns) == 4, "Unexpected number of features inferred"
    assert resp.record_count > 0, "No sample data returned"


def test_infer_schema_from_query_bigquery():
    data_store = get_bigquery_datastore()
    if data_store is None:
        print("Test configuration not found. skipping ")
        pytest.skip("Test configuration not found. skipping")

    fstore = get_featurestore(data_store, projectid="testproject")
    fstore.create_project_schema()

    env = management_types_pb2.Environment(
        name="projects/testproject", data_store=data_store
    )
    req = dataingest_pb2.GetInferredSchemaRequest(
        environment=env,
        query="SELECT id, CAST(id as STRING) as strid, text, rating  from  featurestore_test.tweets order by id",
    )

    os.environ["CONFIG_ENV"] = "local"
    server = DataIngestService()

    resp = server.GetInferredSchema(req, None)

    print("number of features found ", str(len(resp.columns)))
    print("sample data size", resp.record_count)

    assert len(resp.columns) == 4, "Unexpected number of features inferred"
    assert resp.record_count > 0, "No sample data returned"

    req = dataingest_pb2.GetInferredSchemaRequest(
        environment=env,
        query="SELECT id, CAST(id as STRING) as strid, text, rating  from  featurestore_test.tweets order by id",
    )

    resp = server.GetInferredSchema(req, None)

    assert len(resp.columns) == 4, "Unexpected number of features inferred"
    assert resp.record_count > 0, "No sample data returned"


def test_infer_schema_from_query_redshift():
    data_store = get_redshift_datastore()
    if data_store is None:
        print("Test configuration not found. skipping ")
        pytest.skip("Test configuration not found. skipping")

    fstore = get_featurestore(data_store, projectid="testproject")
    fstore.create_project_schema()

    env = management_types_pb2.Environment(
        name="projects/testproject", data_store=data_store
    )
    req = dataingest_pb2.GetInferredSchemaRequest(
        environment=env,
        query="SELECT id, CAST(id as char) as idstr, text, rating  from  featurestore_test.tweets order by id",
    )

    os.environ["CONFIG_ENV"] = "local"
    server = DataIngestService()

    resp = server.GetInferredSchema(req, None)

    print("number of features found ", str(len(resp.columns)))
    print("sample data size", resp.record_count)

    assert len(resp.columns) == 4, "Unexpected number of features inferred"
    assert resp.record_count > 0, "No sample data returned"

    req = dataingest_pb2.GetInferredSchemaRequest(
        environment=env,
        query="SELECT id,  CAST(id as char) as idstr, text, rating  from  featurestore_test.tweets order by id",
    )

    resp = server.GetInferredSchema(req, None)

    assert len(resp.columns) == 4, "Unexpected number of features inferred"
    assert resp.record_count > 0, "No sample data returned"


def test_infer_schema_from_query_databricks():
    data_store = get_databricks_datastore()
    if data_store is None:
        print("Test configuration not found. skipping ")
        pytest.skip("Test configuration not found. skipping")

    fstore = get_featurestore(data_store, projectid="testproject")
    if not fstore.data_store.databricks.db_schema:
        fstore.db_schema = "testproject"
    fstore.create_project_schema()

    env = management_types_pb2.Environment(
        name="projects/testproject", data_store=data_store
    )
    req = dataingest_pb2.GetInferredSchemaRequest(
        environment=env,
        query="SELECT id, CAST(id as string) as idstr, text, rating  from  featurestore_test.tweets order by id",
    )

    os.environ["CONFIG_ENV"] = "local"
    server = DataIngestService()

    resp = server.GetInferredSchema(req, None)

    print("number of features found ", str(len(resp.columns)))
    print("sample data size", resp.record_count)

    assert len(resp.columns) == 4, "Unexpected number of features inferred"
    assert resp.record_count > 0, "No sample data returned"

    req = dataingest_pb2.GetInferredSchemaRequest(
        environment=env,
        query="SELECT id,  CAST(id as string) as idstr, text, rating  from  featurestore_test.tweets order by id",
    )

    resp = server.GetInferredSchema(req, None)

    assert len(resp.columns) == 4, "Unexpected number of features inferred"
    assert resp.record_count > 0, "No sample data returned"



from __future__ import annotations
from typing import List, Optional
from continual.rpc.management.v1 import management_pb2
from continual.rpc.management.v1 import types
from continual.python.sdk.resource import Resource
from continual.python.sdk.manager import Manager
from continual.python.sdk.iterators import Pager
from continual.python.sdk.waiter import wait
from continual.python.sdk.experiments import ExperimentManager, Experiment
from continual.python.sdk.promotions import Promotion
from continual.python.sdk.events import EventManager


class ModelVersionManager(Manager):
    """Manages Model Version resources."""

    name_pattern: str = "projects/{project}/models/{model}/versions/{version}"

    def get(self, id: str) -> ModelVersion:
        """Get model version.

        Arguments:
            id: Model name or id.

        Returns
            An Model Version.
        """
        req = management_pb2.GetModelVersionRequest(name=self.name(id))
        resp = self.client._management.GetModelVersion(req)
        return ModelVersion.from_proto(resp, client=self.client)

    def get_train_config(self, model_version_name: str):
        """Get

        Arguments:
            model_version_name: Model name or id

        Returns
            TrainingJobConfig

        """
        req = management_pb2.GetTrainingJobConfigRequest(
            model_version_name=model_version_name
        )
        resp = self.client._management.GetTrainingJobConfig(req)
        return types.TrainingJobConfig.from_proto(resp, client=self.client)

    def list(
        self,
        page_size: Optional[int] = None,
        filters: List[str] = None,
        all_projects: bool = False,
    ) -> List[ModelVersion]:
        """List model versions.

        Arguments:
            page_size: Number of items to return.

            filter: List of filters to apply to batch prediction job. Can be:
            - state  (i.e. state:FAILED)

        Returns:
            A list of model versions.
        """
        req = management_pb2.ListModelVersionsRequest(
            parent=self.parent,
            page_size=page_size,
            filters=filters,
            all_projects=all_projects,
        )
        resp = self.client._management.ListModelVersions(req)
        return [
            ModelVersion.from_proto(x, client=self.client) for x in resp.model_versions
        ]

    def list_all(self) -> Pager[ModelVersion]:
        """List all model versions.

        Pages through all model versions using an iterator.

        Returns:
            A iterator of all model versions.
        """

        def next_page(next_page_token):
            req = management_pb2.ListModelVersionsRequest(
                parent=self.parent, page_token=next_page_token
            )
            resp = self.client._management.ListModelVersions(req)
            return (
                [
                    ModelVersion.from_proto(x, client=self.client)
                    for x in resp.model_versions
                ],
                resp.next_page_token,
            )

        return Pager(next_page)

    def promote(self, id: str, force: bool = True) -> Promotion:
        """Promote model version.

        Arguments:
            id: Name or id of model version.
            force: Promote even if model performance is worse.
        Returns:
            A new promotion.
        """
        req = management_pb2.PromoteModelVersionRequest(name=self.name(id))
        resp = self.client._management.PromoteModelVersion(req)
        return Promotion.from_proto(resp, client=self.client)

    def cancel(self, id: str) -> None:
        """Cancel model training run."""
        req = management_pb2.CancelTrainingRequest(name=self.name(id))
        self.client._management.CancelTraining(req)


class ModelVersion(Resource, types.ModelVersion):
    """Model version resource."""

    name_pattern: str = "projects/{project}/models/{model}/versions/{version}"
    manager: ModelVersionManager

    experiments: ExperimentManager
    """Experiment manager."""

    events: EventManager
    """Event manager."""

    def _init(self):
        self.manager = ModelVersionManager(parent=self.parent, client=self.client)
        self.experiments = ExperimentManager(parent=self.name, client=self.client)
        self.events = EventManager(parent=self.name, client=self.client)

    def cancel(self) -> None:
        """Cancel model version trailing."""
        self.manager.cancel(self.id)

    def wait(
        self,
        echo: bool = False,
        timeout: Optional[int] = None,
    ) -> ModelVersion:
        """Wait for model state transition to complete.

        Arguments:
            echo: Display progress.
            timeout: Timeout in second.

        Returns:
            An updated model version.
        """
        return wait(lambda: self.manager.get(self.id), echo=echo, timeout=timeout)

    def promote(self, force: bool = True) -> Promotion:
        """Promote model version.

        Arguments:
            force: Promote even if model performance is worse.

        Returns:
            A new promotion.
        """
        return self.manager.promote(self.id, force)

    @property
    def experiment(self) -> Experiment:
        """Winning experiment (expanded object)."""
        return self.client.experiments.get(super().experiment)

    @experiment.setter
    def experiment(self, value: Experiment) -> None:
        types.ModelVersion.experiment.fset(self, value)



from abc import ABC, abstractmethod


class BaseModelPipeline(ABC):
    """
    Abstract Base Class for all pipelines
    """

    @abstractmethod
    def train(self):
        """
        Train model, log a model
        version state, and a model artifact
        """
        pass

    @abstractmethod
    def predict(self):
        """
        Use a trained model to make batch predictions
        on some given data. Log the batch prediction
        state and write the batch predictions to the
        data store
        """
        pass



import typer
import humanize
import os
import yaml

from datetime import datetime
from rich.console import Console
from typing import List

from continual.rpc.management.v1.management_pb2 import YamlDiffRequest
from continual.python.cli import utils
from continual.python.sdk.client import Client
from pathlib import Path

app = typer.Typer(help="Manage changes.")


def truncate(data, n=40):
    """Truncate string with elipses."""
    return (data[:n] + "...") if len(data) > n else data


def format_changes_data(changes, project, environment, n, filters, all_projects):
    data = []
    for push in changes:
        success = 0
        fs_count = 0
        fs_success = 0
        model_count = 0
        model_success = 0
        for step in push.plan:
            if step.resource_name.split("/")[2] == "featureSets":
                fs_count += 1
                if step.state == "SUCCEEDED":
                    fs_success += 1
                    success += 1
            else:
                model_count += 1
                if step.state == "SUCCEEDED":
                    model_success += 1
                    success += 1
        total = len(push.plan)
        age = humanize.naturaltime(datetime.now() - push.create_time)
        push_id = push.name.split("/")[-1]
        push_data = [
            push_id,
            f"{fs_success}/{fs_count}",
            f"{model_success}/{model_count}",
            f"{success}/{total}",
            push.state,
            age,
            truncate(push.message),
        ]
        if all_projects:
            push_data.insert(0, push.parent.split("/")[1])
        data.append(push_data)
    headers = [
        "ID",
        "Feature Set Steps",
        "Model Steps",
        "Total Steps",
        "State",
        "Age",
        "Message",
    ]
    filter_snippet = " (n=%s)" % n
    project_snippet = "project %s, environment %s" % (project, environment)
    if all_projects:
        project_snippet = "all accessible projects & environments"
        headers.insert(0, "Project")
    if len(filters) > 0:
        filter_snippet = " with filter %s" % str(filters) + filter_snippet
    typer.secho(
        "\nFound %s changes in %s%s:" % (len(data), project_snippet, filter_snippet),
        fg="blue",
    )
    return (data, headers)


# use callback to run list command if nothing is passed in
@app.callback(invoke_without_command=True)
def default(ctx: typer.Context):
    if ctx.invoked_subcommand is not None:
        return
    else:
        list(
            project=None,
            environment=None,
            n=30,
            filters=[],
            all_projects=False,
            style=None,
        )


@app.command("list")
@utils.exit_on_error
def list(
    project: str = typer.Option(None, help="Project ID."),
    environment: str = typer.Option(None, "--env", help="Environment ID."),
    n: int = typer.Option(30, "--num", "-n", help="Number of records to show."),
    filters: List[str] = typer.Option([], "--filter", "-f", help="List of filters."),
    all_projects: bool = typer.Option(False, "--all", "-a", help="Show all projects."),
    style: utils.ContinualStyle = typer.Option(None, help="Color to use for list."),
):
    """List all changes.

    Filters can include:
        --state (i.e. state:FAILED)
    """
    project, environment = utils.get_project_and_environment(project, environment)
    c = Client(project=utils.get_environment_name(project, environment))

    push_list = c.changes.list(n, filters=filters, all_projects=all_projects)
    pushes = sorted(push_list, key=lambda x: x.create_time, reverse=True)
    (data, headers) = format_changes_data(
        pushes, project, environment, n, filters, all_projects
    )
    utils.print_table(data, headers, style=utils.get_style(style))


@app.command("get")
@utils.exit_on_error
def get(
    push_id: str = typer.Argument(..., help="Change ID."),
    project: str = typer.Option(None, help="Project ID."),
    environment: str = typer.Option(None, "--env", help="Environment ID."),
    show_diffs: bool = typer.Option(
        True,
        "--skip-diffs",
        help="Don't show diffs between changes and previous schema.",
    ),
    json: bool = typer.Option(False, "--json", help="Show full JSON representation"),
):
    """Get change details."""
    project, environment = utils.get_project_and_environment(project, environment)
    c = Client(project=utils.get_environment_name(project, environment))
    try:
        push = c.changes.get(push_id)
    except Exception as e:
        typer.secho(
            "Failed to get change %s in project %s, environment %s. Please double check the chagne id, project id, and environment id: %s"
            % (push_id, project, environment, str(e)),
            fg="red",
        )
        raise typer.Exit(code=1)
    if json:
        console = Console()
        console.print(push.to_dict())
    else:
        utils.print_change_plan(push, project, environment)
    if show_diffs:
        typer.secho(
            "\nAnalyzing schemas in change plan with previous schemas...", fg="blue"
        )
        for step in push.plan:
            if len(step.schema_text) > 0:
                name = step.resource_name.split("/")[-1]
                previous_schema_text = step.previous_schema_text
                # dyff errors if previous schema is null, so provide a trivial element
                if len(previous_schema_text) == 0:
                    previous_schema_text = "type:"
                req = YamlDiffRequest(
                    right=name,
                    left="%s (previous)" % name,
                    right_text=step.schema_text,
                    left_text=previous_schema_text,
                    parent=push.parent,
                )
                resp = c._management.GetYamlDiff(req)
                diff_text = resp.text
                typer.secho(
                    "Diffing \n\tleft = %s with \n\tright = %s"
                    % ("%s (previous)" % name, name),
                    fg="blue",
                )
                if len(diff_text) > 0:
                    typer.secho(diff_text)
                else:
                    typer.secho("No changes found.", fg="blue")


@app.command("rerun")
@utils.exit_on_error
def rerun(
    push_id: str = typer.Argument(..., help="Change ID."),
    project: str = typer.Option(None, help="Project ID."),
    environment: str = typer.Option(None, "--env", help="Environment ID."),
    message: str = typer.Option("Re-running from CLI", help="Message for new change"),
):
    """Rerun a previous change."""
    project, environment = utils.get_project_and_environment(project, environment)
    c = Client(project=utils.get_environment_name(project, environment))
    try:
        push = c.changes.rerun(push_id)
        utils.print_change(push, project, environment, message)
    except Exception as e:
        typer.secho(
            "Failed to re-run change %s in project %s, environment %s. Please double check the chagne id, project id, and environment id: %s"
            % (push_id, project, environment, str(e)),
            fg="red",
        )


@app.command("cancel")
@utils.exit_on_error
def cancel(
    push_id: str = typer.Argument(..., help="Change ID."),
    project: str = typer.Option(None, help="Project ID."),
    environment: str = typer.Option(None, "--env", help="Environment ID."),
    force: bool = typer.Option(
        False, "--force", help="Force cancellation. Skips confirmation."
    ),
):
    """Cancel change."""
    project, environment = utils.get_project_and_environment(project, environment)
    c = Client(project=utils.get_environment_name(project, environment))
    try:
        if not force:
            force = typer.confirm(
                "Are you sure you want to cancel the change %s?" % push_id
            )
        if force:
            c.changes.cancel(push_id)
            typer.secho("Change %s successfully cancelled." % push_id, fg="green")
    except Exception as e:
        typer.secho(
            "Failed to cancel change %s in project %s, environment %s. Please double check the chagne id, project id, and environment id: %s"
            % (push_id, project, environment, str(e)),
            fg="red",
        )


@app.command("export")
@utils.exit_on_error
def export(
    push_id: str = typer.Argument(..., help="Change ID."),
    project: str = typer.Option(None, "--project", help="Project ID."),
    environment: str = typer.Option(None, "--env", help="Environment ID."),
    path: Path = typer.Option(
        None,
        help="Directory to save projects resources into. Defaults to ./<project_id> if not provided.",
    ),
    # include_predictions: bool = typer.Option(False, "--predictions", hidden=True, help="Include csv of latest predictions for each model"),
):
    """
    Exports change YAMLs to a directory.
    """
    project, environment = utils.get_project_and_environment(project, environment)
    c = Client(project=utils.get_environment_name(project, environment))

    change = c.changes.get(push_id)
    export_change(path, project, change)


def export_change(path, project, change):

    # use cwd/project_id if no path provided
    if path is None:
        path = os.path.join(os.getcwd(), project)
        if not os.path.exists(path):
            os.mkdir(path)

    fs_dir = os.path.join(path, "feature_sets")
    if not os.path.exists(fs_dir):
        os.mkdir(fs_dir)
    model_dir = os.path.join(path, "models")
    if not os.path.exists(model_dir):
        os.mkdir(model_dir)
    # bpj_dir = os.path.join(path,"predictions")

    try:
        fs_yamls = []
        m_yamls = []
        if len(change.plan) > 0:
            for step in change.plan:
                plan_schema = step.schema_text
                if len(plan_schema) > 0:
                    schema_text = yaml.load(plan_schema, Loader=yaml.FullLoader)
                    schema_yaml = utils.format_yaml(schema_text)
                    parts = step.resource_name.split("/")
                    type = parts[2]
                    if type == "featureSets":
                        m_yamls.append((parts[-1], schema_yaml))
                    if type == "models":
                        fs_yamls.append((parts[-1], schema_yaml))

            if len(fs_yamls) > 0:
                utils.save_yamls(fs_yamls, fs_dir)
            if len(m_yamls) > 0:
                utils.save_yamls(m_yamls, model_dir)
            typer.secho(
                "Successfully saved change yamls to directory: %s \n" % fs_dir,
                fg="green",
            )
        else:
            typer.secho("No steps found in change %s" % change.id, fg="blue")
    except Exception as e:
        typer.secho(
            "Failed to get all save change yamls:%s \n" % (str(e)),
            fg="red",
        )



from functools import wraps
import grpc
from grpc_status import rpc_status
from continual.rpc.rpc import error_details_pb2
from typing import TypeVar, Optional, Dict


def map_to_dict(msg) -> dict:
    """Converts SalarMapContainer to dict."""
    return dict(msg.items())


class BaseException(Exception):
    """Base exception for Continual APIs.

    Wraps underlying API errors.
    """

    message: str
    """Error message."""

    details: Dict[str, str]
    """Error details dictionary.
    
    Details includes things like field level validation errors.
    """

    def __init__(self, message: str = "", details: Optional[Dict[str, str]] = None):
        self.message = message
        self.details = map_to_dict(details or dict())
        super().__init__(message, details)


class OtherError(BaseException):
    """Unknown error status."""


class InternalError(BaseException):
    """Internal server error.

    Please contact support if this error persists.
    """


class InvalidArgumentError(BaseException):
    """Invalid argument error.

    Details includes field level error messages.
    """


class AlreadyExistsError(BaseException):
    """Already exists error.

    The resource cannot be created because it already exists.
    """


class NotFoundError(BaseException):
    """Not found error.

    The resource does not exist.
    """


# PermissionDenied means the action was denied due to permissions error.
class PermissionDeniedError(BaseException):
    """Permission denied error.

    The caller does not have the necessary permissions to take an action.
    """


# Unauthenticated means the request does not have valid authentication information.
class UnauthenticatedError(BaseException):
    """Unauthenticated error.

    The caller is not authenticated.
    """


class FailedError(BaseException):
    """Failed error.

    A long-running operation entered a failed state.
    """


class UnsupportedError(BaseException):
    """Unsupported error.

    An operation or field is not supported.
    """


class FailedPreconditionError(BaseException):
    """Failed precondition error.

    A precondition for this operation was not satisfied.
    """


def normalize_exceptions(func):
    """Function decorator for catching common errors and
    re-raising as Continual exception."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except grpc.RpcError as err:
            info = error_details_pb2.ErrorDetails()
            status = rpc_status.from_call(err)
            if (
                status is not None
                and status.details is not None
                and len(status.details) > 0
            ):
                detail = status.details[0]
                if detail.Is(error_details_pb2.ErrorDetails.DESCRIPTOR):
                    detail.Unpack(info)
            msg = err.details()

            if err.code() == grpc.StatusCode.INTERNAL:
                raise InternalError(msg, info.details) from None
            elif err.code() == grpc.StatusCode.INVALID_ARGUMENT:
                raise InvalidArgumentError(msg, info.details) from None
            elif err.code() == grpc.StatusCode.ALREADY_EXISTS:
                raise AlreadyExistsError(msg, info.details) from None
            elif err.code() == grpc.StatusCode.NOT_FOUND:
                raise NotFoundError(msg, info.details) from None
            elif err.code() == grpc.StatusCode.PERMISSION_DENIED:
                raise PermissionDeniedError(msg, info.details)
            elif err.code() == grpc.StatusCode.UNAUTHENTICATED:
                raise UnauthenticatedError(msg, info.details) from None
            elif err.code() == grpc.StatusCode.FAILED_PRECONDITION:
                raise FailedPreconditionError(msg, info.details) from None
            else:
                raise OtherError(msg, info.details) from None

    return wrapper


T = TypeVar("T")


def normalize_exceptions_for_class(x: T) -> T:
    """Wraps all methods in a class instance with an to exception handler.

    This is used to wrap raw GRPC API stubs and normalize all exceptions.
    """
    method_list = [func for func in dir(x) if callable(getattr(x, func))]
    for method in method_list:
        if not method.startswith("__"):
            setattr(x, method, normalize_exceptions(getattr(x, method)))
    return x



from typing import Any
import shortuuid

from continual.python.sdk.extensions.models.base_model_pipeline import BaseModelPipeline
from continual.python.sdk.datasets import Dataset
from continual.python.sdk.client import Client
from continual.rpc.management.v1 import types, management_types_pb2
from continual.python.utils.utils import get_experiment_name, get_current_time


class MyModelPipeline(BaseModelPipeline):
    def train(
        self,
        client: Client,
        model_version_name: str,
        train_data: Dataset,
        output_dir: str,
        fit_config: dict,
        evaluation_model_version_name: str,
        evaluation_model_signature: str,
        model,
        tag,
        data_store,
        log_level: str,
        **kwargs: dict,
    ):
        """
        Train model, log a model
        version state, and a model artifact
        """
        experiment_name = get_experiment_name(
            model_version_name=model_version_name,
            experiment_id=shortuuid.uuid().lower(),
        )
        # Log an experiment
        client.logger.log_field(field_name="start_time", value=get_current_time())
        client.logger.log_experiment(
            types.Experiment(
                name=experiment_name,
                type="TemplateAlgorithm",
                performance_metric="rmse",
                performance_metric_val=40.0,
                state=types.ExperimentState.SUCCEEDED,
                start_time=get_current_time().ToDatetime(),
                finish_time=get_current_time().ToDatetime(),
            )
        )

        # Register winning experiment
        client.logger.log_field(
            "experiment",
            experiment_name,
        )
        client.logger.log_field("performance_metric_val", 40.0)

        # RegisterEvaluation state
        client.logger.log_eval_model_version_metrics(
            types.EvaluationModelVersionMetrics(
                state=types.EvaluationModelVersionMetricsState.NO_DEPLOYED_VERSION
            ).to_proto(),
        )
        client.logger.log_state(state=types.ModelVersionState.SUCCEEDED)

    def predict(
        self,
        client: Client,
        model_version_name: str,
        batch_prediction_name: str,
        predict_data: Dataset,
        model,
        data_store,
        model_version_signature: Any,
        disable_feature_timestamp_generation: bool,
        log_level: str,
        **kwargs: dict,
    ):
        """
        Use a trained model to make batch predictions
        on some given data. Log the batch prediction
        state and write the batch predictions to the
        data store
        """

        client.logger.log_field(field_name="start_time", value=get_current_time())
        ds_stats = management_types_pb2.DatasetStats(
            resource_name=batch_prediction_name,
            profile_time=get_current_time(),
            dataset_stats=[
                management_types_pb2.DatasetStatsEntry(**{}),
            ],
        )
        client.logger.log_dataset_stats(ds_stats)
        client.logger.log_predict_state(state=types.BatchPredictionState.SUCCEEDED)



import re

from databricks import sql
from pyhive.sqlalchemy_hive import HiveDialect
from pyhive.sqlalchemy_hive import _type_map
from sqlalchemy import types
from sqlalchemy import util
from sqlalchemy import exc


class DatabricksDialect(HiveDialect):
    """
    Credit: https://github.com/crflynn/sqlalchemy-databricks
    License: MIT
    """

    name = "databricks"
    driver = "connector"  # databricks-sql-connector

    @classmethod
    def dbapi(cls):
        return sql

    def create_connect_args(self, url):
        # databricks-sql-connector expects just
        # server_hostname, access_token, and http_path
        kwargs = {
            "server_hostname": url.host,
            "access_token": url.password,
        }

        if url.query is not None and "http_path" in url.query:
            kwargs["http_path"] = url.query["http_path"]

        kwargs.update(url.query)
        return [], kwargs

    def get_table_names(self, connection, schema=None, **kw):
        # override to use row[1] in databricks instead of row[0] in hive
        query = "SHOW TABLES"
        if schema:
            query += " IN " + self.identifier_preparer.quote_identifier(schema)
        return [row[1] for row in connection.execute(query)]

    def _get_table_columns(self, connection, table_name, schema):
        full_table = table_name
        if schema:
            full_table = schema + "." + table_name
        # TODO using TGetColumnsReq hangs after sending TFetchResultsReq.
        # Using DESCRIBE works but is uglier.
        try:
            # This needs the table name to be unescaped (no backticks).
            rows = connection.execute("DESCRIBE {}".format(full_table)).fetchall()
        except exc.OperationalError as e:
            # Does the table exist?
            regex_fmt = r"TExecuteStatementResp.*SemanticException.*Table not found {}"
            regex = regex_fmt.format(re.escape(full_table))
            if re.search(regex, e.args[0]):
                raise exc.NoSuchTableError(full_table)
            elif hasattr(e, "orig") and e.orig:
                regex = r"Table or view not found"
                if re.search(regex, str(e.orig)):
                    raise exc.NoSuchTableError(full_table)
                raise

        else:
            # Hive is stupid: this is what I get from DESCRIBE some_schema.does_not_exist
            regex = r"Table .* does not exist"
            if len(rows) == 1 and re.match(regex, rows[0].col_name):
                raise exc.NoSuchTableError(full_table)
            return rows

    def get_columns(self, connection, table_name, schema=None, **kw):
        # override to get columns properly; the reason is because databricks
        # presents the partition information differently from oss hive
        rows = self._get_table_columns(connection, table_name, schema)
        # Strip whitespace
        rows = [[col.strip() if col else None for col in row] for row in rows]
        # Filter out empty rows and comment
        rows = [row for row in rows if row[0] and row[0] != "# col_name"]
        result = []
        for (col_name, col_type, _comment) in rows:
            # Handle both oss hive and Databricks' hive partition header, respectively
            if col_name in ("# Partition Information", "# Partitioning"):
                break
            # Take out the more detailed type information
            # e.g. 'map<int,int>' -> 'map'
            #      'decimal(10,1)' -> decimal
            col_type = re.search(r"^\w+", col_type).group(0)
            try:
                coltype = _type_map[col_type]
            except KeyError:
                util.warn(
                    "Did not recognize type '%s' of column '%s'" % (col_type, col_name)
                )
                coltype = types.NullType

            result.append(
                {
                    "name": col_name,
                    "type": coltype,
                    "nullable": True,
                    "default": None,
                }
            )
        return result


from sqlalchemy.dialects import registry

registry.register("databricks.connector", __name__, DatabricksDialect.__name__)



import unittest
import tempfile
import os
import mock

from continual.python.cli import utils_dbt


class AStringHaving(str):
    def __eq__(self, other):
        return self in other


class process_dbt_project_file(unittest.TestCase):
    def test_should_parse_dbt_project_config(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            dbt_project_file_path = os.path.join(tmpdir, "dbt_project.yml")
            with open(dbt_project_file_path, "w+") as dbt_project_file:
                dbt_project_file.writelines(
                    """
name: 'my_new_project'
version: '1.0.0'
config-version: 2

profile: 'default'

model-paths: ["models"]
analysis-paths: ["analysis"]
test-paths: ["tests"]
seed-paths: ["data"]
macro-paths: ["macros"]
snapshot-paths: ["snapshots"]

target-path: "target"
clean-targets:
  - "target"
  - "dbt_modules"

models:
  my_new_project:                
                """
                )

            (
                dbt_project_config_file_path,
                dbt_project_name,
                dbt_default_profile_name,
                dbt_target_dir,
                dbt_manifest_file_path,
            ) = utils_dbt.process_dbt_project_file(tmpdir)

            assert dbt_project_config_file_path == os.path.join(
                tmpdir, "dbt_project.yml"
            )
            assert dbt_project_name == "my_new_project"
            assert dbt_default_profile_name == "default"
            assert dbt_target_dir == "target"
            assert dbt_manifest_file_path == os.path.join(
                tmpdir, "target/manifest.json"
            )


class get_default_target(unittest.TestCase):
    def test_should_get_profile_target_name(self):
        dbt_profiles_file_path = ""
        with tempfile.TemporaryDirectory() as tmpdir:
            dbt_profiles_file_path = os.path.join(tmpdir, "profiles.yml")
            with open(dbt_profiles_file_path, "w+") as dbt_profiles_file:
                dbt_profiles_file.writelines(
                    """
a_profile_name:
  outputs:
    a_target_name:
      type: snowflake
      account: a_snowflake_account
      user: a_snowflake_user
      password: a_snowflake_password
      role: a_snowflake_role
      database: a_snowflake_database
      warehouse: a_snowflake_warehouse
      schema: a_snowflake_schema_name

  target: a_target_name            
                """
                )

            result = utils_dbt.get_default_target(
                dbt_profiles_file_path, "a_profile_name"
            )
            assert result == "a_target_name"

    def test_should_handle_case_sensitive_profile_names(self):
        dbt_profiles_file_path = ""
        with tempfile.TemporaryDirectory() as tmpdir:
            dbt_profiles_file_path = os.path.join(tmpdir, "profiles.yml")
            with open(dbt_profiles_file_path, "w+") as dbt_profiles_file:
                dbt_profiles_file.writelines(
                    """
A_Profile_Name_With_Capital_Letters:
  outputs:
    a_target_name:
      type: snowflake
      account: a_snowflake_account
      user: a_snowflake_user
      password: a_snowflake_password
      role: a_snowflake_role
      database: a_snowflake_database
      warehouse: a_snowflake_warehouse
      schema: a_snowflake_schema_name

  target: a_target_name
                """
                )

            result = utils_dbt.get_default_target(
                dbt_profiles_file_path, "A_Profile_Name_With_Capital_Letters"
            )
            assert result == "a_target_name"

    def test_should_handle_case_sensitive_keys(self):
        dbt_profiles_file_path = ""
        with tempfile.TemporaryDirectory() as tmpdir:
            dbt_profiles_file_path = os.path.join(tmpdir, "profiles.yml")
            with open(dbt_profiles_file_path, "w+") as dbt_profiles_file:
                dbt_profiles_file.writelines(
                    """
a_profile_name:
  outputs:
    a_target_name:
      type: snowflake
      account: a_snowflake_account
      user: a_snowflake_user
      password: a_snowflake_password
      role: a_snowflake_role
      database: a_snowflake_database
      warehouse: a_snowflake_warehouse
      schema: a_snowflake_schema_name

  TARGET: a_target_name     
                """
                )

            result = utils_dbt.get_default_target(
                dbt_profiles_file_path, "a_profile_name"
            )
            # dbt requires keys to be lowercase: https://linear.app/continual/issue/DEV-2805/dbt-profile-config-keys-are-case-sensitive
            assert result is None


class get_default_target(unittest.TestCase):
    def test_should_get_default_target_for_profile(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            dbt_profiles_file_path = os.path.join(tmpdir, "dbt_profiles.yml")
            with open(dbt_profiles_file_path, "w+") as dbt_profiles_file:
                dbt_profiles_file.writelines(
                    """
a_profile_name:
  outputs:
    a_target:
      type: snowflake
      account: some_account
      user: some_user
      password: some_password
      role: some_role
      database: some_database
      warehouse: some_warehouse
      schema: some_schema

  target: a_target             
                """
                )

            dbt_target_result = utils_dbt.get_default_target(
                dbt_profiles_file_path, "a_profile_name"
            )

            assert dbt_target_result == "a_target"

    def test_should_fail_when_default_target_is_not_defined(self):
        # Note that this misconfiguration also causes `dbt debug` to fail
        with tempfile.TemporaryDirectory() as tmpdir:
            dbt_profiles_file_path = os.path.join(tmpdir, "dbt_profiles.yml")
            with open(dbt_profiles_file_path, "w+") as dbt_profiles_file:
                dbt_profiles_file.writelines(
                    """
a_profile_name:
  outputs:
    a_different_target:
      type: snowflake
    another_different_target:
      type: snowflake

  target: a_target_without_a_definition
                """
                )

            with mock.patch("typer.secho") as secho:
                with self.assertRaises(SystemExit) as e:
                    utils_dbt.get_default_target(
                        dbt_profiles_file_path, "a_profile_name"
                    )
                secho.assert_any_call(
                    AStringHaving("Default dbt target definition not found"),
                    fg=mock.ANY,
                )


class dbt_target_name_is_defined(unittest.TestCase):
    def test_should_return_true_when_target_is_defined(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            dbt_profiles_file_path = os.path.join(tmpdir, "dbt_profiles.yml")
            with open(dbt_profiles_file_path, "w+") as dbt_profiles_file:
                dbt_profiles_file.writelines(
                    """
a_profile_name:
  outputs:
    a_target:
      type: snowflake
      account: some_account
      user: some_user
      password: some_password
      role: some_role
      database: some_database
      warehouse: some_warehouse
      schema: some_schema

  target: a_target             
                """
                )

            result = utils_dbt.dbt_target_name_is_defined(
                dbt_profiles_file_path, "a_profile_name", "a_target"
            )
            assert result == True

    def test_should_return_false_and_secho_when_target_is_not_defined(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            dbt_profiles_file_path = os.path.join(tmpdir, "dbt_profiles.yml")
            with open(dbt_profiles_file_path, "w+") as dbt_profiles_file:
                dbt_profiles_file.writelines(
                    """
a_profile_name:
  outputs:
    a_target:
      type: snowflake
      account: some_account
      user: some_user
      password: some_password
      role: some_role
      database: some_database
      warehouse: some_warehouse
      schema: some_schema

  target: a_target
                """
                )

            with mock.patch("typer.secho") as secho:
                result = utils_dbt.dbt_target_name_is_defined(
                    dbt_profiles_file_path, "a_profile_name", "a_non_existent_target"
                )
                assert result == False
                secho.assert_any_call(
                    AStringHaving(
                        "dbt target definition not found for target 'a_non_existent_target'."
                    ),
                    fg=mock.ANY,
                )


class dbt_profile_name_is_defined(unittest.TestCase):
    def test_should_return_true_when_profile_is_defined(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            dbt_profiles_file_path = os.path.join(tmpdir, "dbt_profiles.yml")
            with open(dbt_profiles_file_path, "w+") as dbt_profiles_file:
                dbt_profiles_file.writelines(
                    """
a_profile_name:
  outputs:
    a_target:
      type: snowflake
      account: some_account
      user: some_user
      password: some_password
      role: some_role
      database: some_database
      warehouse: some_warehouse
      schema: some_schema

  target: a_target             
                """
                )

            result = utils_dbt.dbt_profile_name_is_defined(
                dbt_profiles_file_path, "a_profile_name"
            )
            assert result == True

    def test_should_return_false_and_secho_when_profile_is_not_defined(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            dbt_profiles_file_path = os.path.join(tmpdir, "dbt_profiles.yml")
            with open(dbt_profiles_file_path, "w+") as dbt_profiles_file:
                dbt_profiles_file.writelines(
                    """
a_profile_name:
  outputs:
    a_target:
      type: snowflake
      account: some_account
      user: some_user
      password: some_password
      role: some_role
      database: some_database
      warehouse: some_warehouse
      schema: some_schema

  target: a_target
                """
                )

            with mock.patch("typer.secho") as secho:
                result = utils_dbt.dbt_profile_name_is_defined(
                    dbt_profiles_file_path, "a_non_existent_profile_name"
                )
                assert result == False
                secho.assert_any_call(
                    AStringHaving(
                        "dbt profile definition not found for profile 'a_non_existent_profile_name'."
                    ),
                    fg=mock.ANY,
                )


class dbt_target_name_as_environment_id(unittest.TestCase):
    def test_should_replace_dashes_with_underscores(self):
        assert (
            utils_dbt.dbt_target_name_as_environment_id("target-name-with-dashes")
            == "target_name_with_dashes"
        )


class is_dbt_project_dir(unittest.TestCase):
    def test_should_return_true_when_dbt_config_file_exists(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            dbt_profiles_file_path = os.path.join(tmpdir, "dbt_project.yml")
            with open(dbt_profiles_file_path, "w+") as dbt_profiles_file:
                dbt_profiles_file.writelines(
                    """
name: 'my_dbt_project'
                    """
                )

            result = utils_dbt.is_dbt_project_dir(tmpdir)
            assert result == True

    def test_should_return_false_when_dbt_config_file_does_not_exist(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            result = utils_dbt.is_dbt_project_dir(tmpdir)
            assert result == False
